#    Copyright 2015 Yahoo Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinder import db
from cinder import exception
from cinder.i18n import _
from cinder import objects
from cinder.objects import base
from cinder.objects import fields as c_fields
from oslo_versionedobjects import fields

OPTIONAL_FIELDS = ['cgsnapshots', 'volumes']


@base.CinderObjectRegistry.register
class ConsistencyGroup(base.CinderPersistentObject, base.CinderObject,
                       base.CinderObjectDictCompat):
    # Version 1.0: Initial version
    # Version 1.1: Added cgsnapshots and volumes relationships
    # Version 1.2: Changed 'status' field to use ConsistencyGroupStatusField
    VERSION = '1.2'

    fields = {
        'id': fields.UUIDField(),
        'user_id': fields.UUIDField(),
        'project_id': fields.UUIDField(),
        'host': fields.StringField(nullable=True),
        'availability_zone': fields.StringField(nullable=True),
        'name': fields.StringField(nullable=True),
        'description': fields.StringField(nullable=True),
        'volume_type_id': fields.UUIDField(nullable=True),
        'status': c_fields.ConsistencyGroupStatusField(nullable=True),
        'cgsnapshot_id': fields.UUIDField(nullable=True),
        'source_cgid': fields.UUIDField(nullable=True),
        'cgsnapshots': fields.ObjectField('CGSnapshotList', nullable=True),
        'volumes': fields.ObjectField('VolumeList', nullable=True),
    }

    @staticmethod
    def _from_db_object(context, consistencygroup, db_consistencygroup,
                        expected_attrs=None):
        if expected_attrs is None:
            expected_attrs = []
        for name, field in consistencygroup.fields.items():
            if name in OPTIONAL_FIELDS:
                continue
            value = db_consistencygroup.get(name)
            setattr(consistencygroup, name, value)

        if 'cgsnapshots' in expected_attrs:
            cgsnapshots = base.obj_make_list(
                context, objects.CGSnapshotsList(context),
                objects.CGSnapshot,
                db_consistencygroup['cgsnapshots'])
            consistencygroup.cgsnapshots = cgsnapshots

        if 'volumes' in expected_attrs:
            volumes = base.obj_make_list(
                context, objects.VolumeList(context),
                objects.Volume,
                db_consistencygroup['volumes'])
            consistencygroup.cgsnapshots = volumes

        consistencygroup._context = context
        consistencygroup.obj_reset_changes()
        return consistencygroup

    @base.remotable
    def create(self):
        if self.obj_attr_is_set('id'):
            raise exception.ObjectActionError(action='create',
                                              reason=_('already_created'))
        updates = self.cinder_obj_get_changes()

        if 'cgsnapshots' in updates:
            raise exception.ObjectActionError(action='create',
                                              reason=_('cgsnapshots assigned'))

        if 'volumes' in updates:
            raise exception.ObjectActionError(action='create',
                                              reason=_('volumes assigned'))

        db_consistencygroups = db.consistencygroup_create(self._context,
                                                          updates)
        self._from_db_object(self._context, self, db_consistencygroups)

    def obj_load_attr(self, attrname):
        if attrname not in OPTIONAL_FIELDS:
            raise exception.ObjectActionError(
                action='obj_load_attr',
                reason=_('attribute %s not lazy-loadable') % attrname)
        if not self._context:
            raise exception.OrphanedObjectError(method='obj_load_attr',
                                                objtype=self.obj_name())

        if attrname == 'cgsnapshots':
            self.cgsnapshots = objects.CGSnapshotList.get_all_by_group(
                self._context, self.id)

        if attrname == 'volumes':
            self.volumes = objects.VolumeList.get_all_by_group(self._context,
                                                               self.id)

        self.obj_reset_changes(fields=[attrname])

    @base.remotable
    def save(self):
        updates = self.cinder_obj_get_changes()
        if updates:
            if 'cgsnapshots' in updates:
                raise exception.ObjectActionError(
                    action='save', reason=_('cgsnapshots changed'))
            if 'volumes' in updates:
                raise exception.ObjectActionError(
                    action='save', reason=_('volumes changed'))

            db.consistencygroup_update(self._context, self.id, updates)
            self.obj_reset_changes()

    @base.remotable
    def destroy(self):
        with self.obj_as_admin():
            db.consistencygroup_destroy(self._context, self.id)


@base.CinderObjectRegistry.register
class ConsistencyGroupList(base.ObjectListBase, base.CinderObject):
    # Version 1.0: Initial version
    # Version 1.1: Add pagination support to consistency group
    VERSION = '1.1'

    fields = {
        'objects': fields.ListOfObjectsField('ConsistencyGroup')
    }
    child_version = {
        '1.0': '1.0',
        '1.1': '1.1',
    }

    @base.remotable_classmethod
    def get_all(cls, context, filters=None, marker=None, limit=None,
                offset=None, sort_keys=None, sort_dirs=None):
        consistencygroups = db.consistencygroup_get_all(
            context, filters=filters, marker=marker, limit=limit,
            offset=offset, sort_keys=sort_keys, sort_dirs=sort_dirs)
        return base.obj_make_list(context, cls(context),
                                  objects.ConsistencyGroup,
                                  consistencygroups)

    @base.remotable_classmethod
    def get_all_by_project(cls, context, project_id, filters=None, marker=None,
                           limit=None, offset=None, sort_keys=None,
                           sort_dirs=None):
        consistencygroups = db.consistencygroup_get_all_by_project(
            context, project_id, filters=filters, marker=marker, limit=limit,
            offset=offset, sort_keys=sort_keys, sort_dirs=sort_dirs)
        return base.obj_make_list(context, cls(context),
                                  objects.ConsistencyGroup,
                                  consistencygroups)



# Copyright 2013 eBay Inc.
# Copyright 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from cinder.api.contrib import scheduler_stats
from cinder import context
from cinder import test
from cinder.tests.unit.api import fakes


def schedule_rpcapi_get_pools(self, context, filters=None):
    all_pools = []
    pool1 = dict(name='pool1',
                 capabilities=dict(
                     total_capacity=1024, free_capacity=100,
                     volume_backend_name='pool1', reserved_percentage=0,
                     driver_version='1.0.0', storage_protocol='iSCSI',
                     QoS_support='False', updated=None))
    all_pools.append(pool1)
    pool2 = dict(name='pool2',
                 capabilities=dict(
                     total_capacity=512, free_capacity=200,
                     volume_backend_name='pool2', reserved_percentage=0,
                     driver_version='1.0.1', storage_protocol='iSER',
                     QoS_support='True', updated=None))
    all_pools.append(pool2)

    return all_pools


@mock.patch('cinder.scheduler.rpcapi.SchedulerAPI.get_pools',
            schedule_rpcapi_get_pools)
class SchedulerStatsAPITest(test.TestCase):
    def setUp(self):
        super(SchedulerStatsAPITest, self).setUp()
        self.flags(host='fake')
        self.controller = scheduler_stats.SchedulerStatsController()
        self.ctxt = context.RequestContext('admin', 'fake', True)

    def test_get_pools_summery(self):
        req = fakes.HTTPRequest.blank('/v2/fake/scheduler_stats')
        req.environ['cinder.context'] = self.ctxt
        res = self.controller.get_pools(req)

        self.assertEqual(2, len(res['pools']))

        expected = {
            'pools': [
                {
                    'name': 'pool1',
                },
                {
                    'name': 'pool2',
                }
            ]
        }

        self.assertDictMatch(expected, res)

    def test_get_pools_detail(self):
        req = fakes.HTTPRequest.blank('/v2/fake/scheduler_stats?detail=True')
        req.environ['cinder.context'] = self.ctxt
        res = self.controller.get_pools(req)

        self.assertEqual(2, len(res['pools']))

        expected = {
            'pools': [
                {
                    'name': 'pool1',
                    'capabilities': {
                        'updated': None,
                        'total_capacity': 1024,
                        'free_capacity': 100,
                        'volume_backend_name': 'pool1',
                        'reserved_percentage': 0,
                        'driver_version': '1.0.0',
                        'storage_protocol': 'iSCSI',
                        'QoS_support': 'False', }
                },
                {
                    'name': 'pool2',
                    'capabilities': {
                        'updated': None,
                        'total_capacity': 512,
                        'free_capacity': 200,
                        'volume_backend_name': 'pool2',
                        'reserved_percentage': 0,
                        'driver_version': '1.0.1',
                        'storage_protocol': 'iSER',
                        'QoS_support': 'True', }
                }
            ]
        }

        self.assertDictMatch(expected, res)



# Copyright 2016 Nexenta Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
:mod:`nexenta.nfs` -- Driver to store volumes on NexentaStor Appliance.
=======================================================================

.. automodule:: nexenta.nfs
"""

import hashlib
import os

from oslo_log import log as logging

from cinder import context
from cinder import db
from cinder import exception
from cinder.i18n import _, _LE, _LI, _LW
from cinder.volume.drivers.nexenta.ns5 import jsonrpc
from cinder.volume.drivers.nexenta import options
from cinder.volume.drivers.nexenta import utils
from cinder.volume.drivers import nfs

VERSION = '1.0.0'
LOG = logging.getLogger(__name__)


class NexentaNfsDriver(nfs.NfsDriver):  # pylint: disable=R0921
    """Executes volume driver commands on Nexenta Appliance.

    Version history:
        1.0.0 - Initial driver version.
    """

    driver_prefix = 'nexenta'
    volume_backend_name = 'NexentaNfsDriver'
    VERSION = VERSION

    def __init__(self, *args, **kwargs):
        super(NexentaNfsDriver, self).__init__(*args, **kwargs)
        if self.configuration:
            self.configuration.append_config_values(
                options.NEXENTA_CONNECTION_OPTS)
            self.configuration.append_config_values(
                options.NEXENTA_NFS_OPTS)
            self.configuration.append_config_values(
                options.NEXENTA_DATASET_OPTS)

        self.nfs_mount_point_base = self.configuration.nexenta_mount_point_base
        self.dataset_compression = (
            self.configuration.nexenta_dataset_compression)
        self.dataset_deduplication = self.configuration.nexenta_dataset_dedup
        self.dataset_description = (
            self.configuration.nexenta_dataset_description)
        self.sparsed_volumes = self.configuration.nexenta_sparsed_volumes
        self.nef = None
        self.nef_protocol = self.configuration.nexenta_rest_protocol
        self.nef_host = self.configuration.nas_ip
        self.share = self.configuration.nas_share_path
        self.nef_port = self.configuration.nexenta_rest_port
        self.nef_user = self.configuration.nexenta_user
        self.nef_password = self.configuration.nexenta_password

    @property
    def backend_name(self):
        backend_name = None
        if self.configuration:
            backend_name = self.configuration.safe_get('volume_backend_name')
        if not backend_name:
            backend_name = self.__class__.__name__
        return backend_name

    def do_setup(self, context):
        if self.nef_protocol == 'auto':
            protocol, auto = 'http', True
        else:
            protocol, auto = self.nef_protocol, False
        self.nef = jsonrpc.NexentaJSONProxy(
            protocol, self.nef_host, self.nef_port, self.nef_user,
            self.nef_password, auto=auto)

    def check_for_setup_error(self):
        """Verify that the volume for our folder exists.

        :raise: :py:exc:`LookupError`
        """
        pool_name, fs = self._get_share_datasets(self.share)
        url = 'storage/pools/%s' % (pool_name)
        if not self.nef.get(url):
            raise LookupError(_("Pool %s does not exist in Nexenta "
                                "Store appliance") % pool_name)
        url = 'storage/pools/%s/filesystems/%s' % (
            pool_name, fs)
        if not self.nef.get(url):
            raise LookupError(_("filesystem %s does not exist in "
                                "Nexenta Store appliance") % fs)

        path = '/'.join([pool_name, fs])
        shared = False
        response = self.nef.get('nas/nfs')
        for share in response['data']:
            if share.get('filesystem') == path:
                shared = True
                break
        if not shared:
            raise LookupError(_("Dataset %s is not shared in Nexenta "
                                "Store appliance") % path)

    def initialize_connection(self, volume, connector):
        """Allow connection to connector and return connection info.

        :param volume: volume reference
        :param connector: connector reference
        """
        data = {'export': volume['provider_location'], 'name': 'volume'}
        if volume['provider_location'] in self.shares:
            data['options'] = self.shares[volume['provider_location']]
        return {
            'driver_volume_type': self.driver_volume_type,
            'data': data
        }

    def create_volume(self, volume):
        """Creates a volume.

        :param volume: volume reference
        :returns: provider_location update dict for database
        """
        self._do_create_volume(volume)
        return {'provider_location': volume['provider_location']}

    def _do_create_volume(self, volume):
        pool, fs = self._get_share_datasets(self.share)
        filesystem = '%s/%s/%s' % (pool, fs, volume['name'])
        LOG.debug('Creating filesystem on NexentaStor %s', filesystem)
        url = 'storage/pools/%s/filesystems' % pool
        data = {
            'name': '/'.join([fs, volume['name']]),
            'compressionMode': self.dataset_compression,
            'dedupMode': self.dataset_deduplication,
        }
        self.nef.post(url, data)
        volume['provider_location'] = '%s:/%s/%s' % (
            self.nef_host, self.share, volume['name'])
        try:
            self._share_folder(fs, volume['name'])
            self._ensure_share_mounted('%s:/%s/%s' % (
                self.nef_host, self.share, volume['name']))

            volume_size = volume['size']
            if getattr(self.configuration,
                       self.driver_prefix + '_sparsed_volumes'):
                self._create_sparsed_file(self.local_path(volume), volume_size)
            else:
                url = 'storage/pools/%s/filesystems/%s' % (
                    pool, '%2F'.join([fs, volume['name']]))
                compression = self.nef.get(url).get('compressionMode')
                if compression != 'off':
                    # Disable compression, because otherwise will not use space
                    # on disk.
                    self.nef.put(url, {'compressionMode': 'off'})
                try:
                    self._create_regular_file(
                        self.local_path(volume), volume_size)
                finally:
                    if compression != 'off':
                        # Backup default compression value if it was changed.
                        self.nef.put(url, {'compressionMode': compression})

        except exception.NexentaException:
            try:
                url = 'storage/pools/%s/filesystems/%s/%s' % (
                    pool, '%2F'.join([fs, volume['name']]))
                self.nef.delete(url)
            except exception.NexentaException:
                LOG.warning(_LW("Cannot destroy created folder: "
                                "%(vol)s/%(folder)s"),
                            {'vol': pool, 'folder': '/'.join(
                                [fs, volume['name']])})
            raise

    def delete_volume(self, volume):
        """Deletes a logical volume.

        :param volume: volume reference
        """
        pool, fs = self._get_share_datasets(self.share)
        url = ('storage/pools/%(pool)s/filesystems/%(fs)s') % {
            'pool': pool,
            'fs': '%2F'.join([fs, volume['name']])
        }
        origin = self.nef.get(url).get('originalSnapshot')
        url = ('storage/pools/%(pool)s/filesystems/'
               '%(fs)s?snapshots=true') % {
            'pool': pool,
            'fs': '%2F'.join([fs, volume['name']])
        }
        try:
            self.nef.delete(url)
        except exception.NexentaException as exc:
            if 'Failed to destroy snapshot' in exc.args[0]:
                LOG.debug('Snapshot has dependent clones, skipping')
            else:
                raise
        try:
            if origin and self._is_clone_snapshot_name(origin):
                path, snap = origin.split('@')
                pool, fs = path.split('/', 1)
                snap_url = ('storage/pools/%(pool)s/'
                            'filesystems/%(fs)s/snapshots/%(snap)s') % {
                    'pool': pool,
                    'fs': fs,
                    'snap': snap
                }
                self.nef.delete(snap_url)
        except exception.NexentaException as exc:
            if 'does not exist' in exc.args[0]:
                LOG.debug(
                    'Volume %s does not exist on appliance', '/'.join(
                        [pool, fs]))

    def create_snapshot(self, snapshot):
        """Creates a snapshot.

        :param snapshot: snapshot reference
        """
        volume = self._get_snapshot_volume(snapshot)
        pool, fs = self._get_share_datasets(self.share)
        url = 'storage/pools/%(pool)s/filesystems/%(fs)s/snapshots' % {
            'pool': pool,
            'fs': '%2F'.join([fs, volume['name']]),
        }
        data = {'name': snapshot['name']}
        self.nef.post(url, data)

    def delete_snapshot(self, snapshot):
        """Deletes a snapshot.

        :param snapshot: snapshot reference
        """
        volume = self._get_snapshot_volume(snapshot)
        pool, fs = self._get_share_datasets(self.share)
        url = ('storage/pools/%(pool)s/'
               'filesystems/%(fs)s/snapshots/%(snap)s') % {
            'pool': pool,
            'fs': '%2F'.join([fs, volume['name']]),
            'snap': snapshot['name']
        }
        try:
            self.nef.delete(url)
        except exception.NexentaException as exc:
            if 'EBUSY' is exc:
                LOG.warning(_LW(
                    'Could not delete snapshot %s - it has dependencies'),
                    snapshot['name'])

    def create_volume_from_snapshot(self, volume, snapshot):
        """Create new volume from other's snapshot on appliance.

        :param volume: reference of volume to be created
        :param snapshot: reference of source snapshot
        """
        snapshot_vol = self._get_snapshot_volume(snapshot)
        volume['provider_location'] = snapshot_vol['provider_location']

        pool, fs = self._get_share_datasets(self.share)
        dataset_path = '%s/%s' % (pool, fs)
        url = ('storage/pools/%(pool)s/'
               'filesystems/%(fs)s/snapshots/%(snap)s/clone') % {
            'pool': pool,
            'fs': '%2F'.join([fs, snapshot_vol['name']]),
            'snap': snapshot['name']
        }
        path = '/'.join([pool, fs, volume['name']])
        data = {'targetPath': path}
        self.nef.post(url, data)
        path = '%2F'.join([pool, fs, volume['name']])
        url = 'storage/filesystems/%s/promote' % path
        self.nef.post(url)

        try:
            self._share_folder(fs, volume['name'])
        except exception.NexentaException:
            try:
                url = ('storage/pools/%(pool)s/'
                       'filesystems/%(fs)s') % {
                    'pool': pool,
                    'fs': volume['name']
                }
                self.nef.delete(url)
            except exception.NexentaException:
                LOG.warning(_LW("Cannot destroy cloned filesystem: "
                                "%(vol)s/%(filesystem)s"),
                            {'vol': dataset_path,
                            'filesystem': volume['name']})
            raise

        return {'provider_location': volume['provider_location']}

    def create_cloned_volume(self, volume, src_vref):
        """Creates a clone of the specified volume.

        :param volume: new volume reference
        :param src_vref: source volume reference
        """
        LOG.info(_LI('Creating clone of volume: %s'), src_vref['id'])
        snapshot = {'volume_name': src_vref['name'],
                    'volume_id': src_vref['id'],
                    'name': self._get_clone_snapshot_name(volume)}
        self.create_snapshot(snapshot)
        try:
            return self.create_volume_from_snapshot(volume, snapshot)
        except exception.NexentaException:
            LOG.error(_LE('Volume creation failed, deleting created snapshot '
                          '%(volume_name)s@%(name)s'), snapshot)
            try:
                self.delete_snapshot(snapshot)
            except (exception.NexentaException, exception.SnapshotIsBusy):
                LOG.warning(_LW('Failed to delete zfs snapshot '
                                '%(volume_name)s@%(name)s'), snapshot)
            raise
        self.delete_snapshot(snapshot)

    def local_path(self, volume):
        """Get volume path (mounted locally fs path) for given volume.

        :param volume: volume reference
        """
        nfs_share = volume['provider_location']
        return os.path.join(self._get_mount_point_for_share(nfs_share),
                            'volume')

    def _get_mount_point_for_share(self, nfs_share):
        """Returns path to mount point NFS share.

        :param nfs_share: example 172.18.194.100:/var/nfs
        """
        nfs_share = nfs_share.encode('utf-8')
        return os.path.join(self.configuration.nexenta_mount_point_base,
                            hashlib.md5(nfs_share).hexdigest())

    def _share_folder(self, path, filesystem):
        """Share NFS filesystem on NexentaStor Appliance.

        :param nef: nef object
        :param path: path to parent filesystem
        :param filesystem: filesystem that needs to be shared
        """
        pool = self.share.split('/')[0]
        LOG.debug(
            'Creating ACL for filesystem %s on Nexenta Store', filesystem)
        url = 'storage/pools/%s/filesystems/%s/acl' % (
            pool, '%2F'.join([path.replace('/', '%2F'), filesystem]))
        data = {
            "type": "allow",
            "principal": "everyone@",
            "permissions": [
                "list_directory",
                "read_data",
                "add_file",
                "write_data",
                "add_subdirectory",
                "append_data",
                "read_xattr",
                "write_xattr",
                "execute",
                "delete_child",
                "read_attributes",
                "write_attributes",
                "delete",
                "read_acl",
                "write_acl",
                "write_owner",
                "synchronize"
            ],
            "flags": [
                "file_inherit",
                "dir_inherit"
            ]
        }
        self.nef.post(url, data)

        LOG.debug(
            'Successfully shared filesystem %s', '/'.join(
                [path, filesystem]))

    def _get_capacity_info(self, path):
        """Calculate available space on the NFS share.

        :param path: example pool/nfs
        """
        pool, fs = self._get_share_datasets(path)
        url = 'storage/pools/%s/filesystems/%s' % (
            pool, fs)
        data = self.nef.get(url)
        total = utils.str2size(data['bytesAvailable'])
        allocated = utils.str2size(data['bytesUsed'])
        free = total - allocated
        return total, free, allocated

    def _get_snapshot_volume(self, snapshot):
        ctxt = context.get_admin_context()
        return db.volume_get(ctxt, snapshot['volume_id'])

    def _get_share_datasets(self, nfs_share):
        pool_name, fs = nfs_share.split('/', 1)
        return pool_name, fs

    def _get_clone_snapshot_name(self, volume):
        """Return name for snapshot that will be used to clone the volume."""
        return 'cinder-clone-snapshot-%(id)s' % volume

    def _is_clone_snapshot_name(self, snapshot):
        """Check if snapshot is created for cloning."""
        name = snapshot.split('@')[-1]
        return name.startswith('cinder-clone-snapshot-')

    def _update_volume_stats(self):
        """Retrieve stats info for NexentaStor appliance."""
        LOG.debug('Updating volume stats')
        share = ':/'.join([self.nef_host, self.share])
        total, free, allocated = self._get_capacity_info(self.share)
        total_space = utils.str2gib_size(total)
        free_space = utils.str2gib_size(free)

        location_info = '%(driver)s:%(share)s' % {
            'driver': self.__class__.__name__,
            'share': share
        }
        self._stats = {
            'vendor_name': 'Nexenta',
            'dedup': self.dataset_deduplication,
            'compression': self.dataset_compression,
            'description': self.dataset_description,
            'nef_url': self.nef_host,
            'driver_version': self.VERSION,
            'storage_protocol': 'NFS',
            'total_capacity_gb': total_space,
            'free_capacity_gb': free_space,
            'reserved_percentage': self.configuration.reserved_percentage,
            'QoS_support': False,
            'location_info': location_info,
            'volume_backend_name': self.backend_name,
            'nfs_mount_point_base': self.nfs_mount_point_base
        }



# Copyright (c) 2014 Hitachi Data Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
import time

import mock
from oslo_concurrency import processutils as putils
from oslo_config import cfg

from cinder import exception
from cinder import test
from cinder import utils
from cinder.volume.drivers.hitachi import hnas_backend
from cinder.volume.drivers.hitachi import hnas_nfs as nfs

CONF = cfg.CONF

HNAS_RESULT1 = "\n\
FS ID        FS Label        FS Permanent ID     EVS ID     EVS Label\n\
-----     -----------     ------------------     ------     ---------\n\
 1026            gold     0xaadee0e035cfc0b7          1      EVSTest1\n\
 1025      fs01-husvm     0xaada5dff78668800          1      EVSTest1\n\
 1027     large-files     0xaadee0ef012a0d54          1      EVSTest1\n\
 1028        platinun     0xaadee1ea49d1a32c          1      EVSTest1\n\
 1029        test_hdp     0xaadee09634acfcac          1      EVSTest1\n\
 1030         cinder1     0xaadfcf742fba644e          1      EVSTest1\n\
 1031         cinder2     0xaadfcf7e0769a6bc          1      EVSTest1\n\
 1024      fs02-husvm     0xaac8715e2e9406cd          2      EVSTest2\n\
\n"

HNAS_RESULT2 = "cluster MAC: 83-68-96-AA-DA-5D"

HNAS_RESULT3 = "\n\
Model: HNAS 4040                                             \n\
Software: 11.2.3319.14 (built 2013-09-19 12:34:24+01:00)     \n\
Hardware: NAS Platform (M2SEKW1339109)                       \n\
board        MMB1                                            \n\
mmb          11.2.3319.14 release (2013-09-19 12:34:24+01:00)\n\
board        MFB1                                            \n\
mfb1hw       MB v0883 WL v002F TD v002F FD v002F TC v0059      \
    RY v0059 TY v0059 IC v0059 WF v00E2 FS v00E2 OS v00E2      \
    WD v00E2 DI v001A FC v0002                               \n\
Serial no    B1339745 (Thu Jan  1 00:00:50 2009)             \n\
board        MCP                                             \n\
Serial no    B1339109 (Thu Jan  1 00:00:49 2009)             \n\
\n"

HNAS_RESULT4 = "\n\
EVS Type    Label            IP Address          Mask             Port  \n\
----------  ---------------  ------------------  ---------------  ------\n\
admin       hnas4040         192.0.2.2           255.255.255.0    eth1  \n\
admin       hnas4040         172.24.44.15        255.255.255.0    eth0  \n\
evs 1       EVSTest1         172.24.44.20        255.255.255.0    ag1   \n\
evs 1       EVSTest1         10.0.0.20           255.255.255.0    ag1   \n\
evs 2       EVSTest2         172.24.44.21        255.255.255.0    ag1   \n\
\n"

HNAS_RESULT5 = "\n\
 ID         Label  EVS     Size     Used       Snapshots  Deduped\
     Avail        Thin  ThinSize  ThinAvail                      \
            FS Type \n\
----  -----------  ---  ------- -------------  ---------  -------\
- -------------  ----  --------  ---------  ---------------------\
------------- \n\
1025   fs01-husvm    1   250 GB 21.4 GB (9%)  0 B (0%)   NA      \
   228 GB (91%)   No                                       32 KB,\
   WFS-2,128 DSBs\n\
1026         gold    1  19.9 GB 2.30 GB (12%    NA       0 B (0%)\
   17.6 GB (88%)  No                         4 KB,WFS-2,128 DSBs,\
   dedupe enabled\n\
1027  large-files    1  19.8 GB 2.43 GB (12%) 0 B (0%)   NA      \
   17.3 GB (88%)  No                                       32 KB,\
   WFS-2,128 DSBs\n\
1028     platinun    1  19.9 GB 2.30 GB (12%)   NA       0 B (0%)\
   17.6 GB (88%)  No                         4 KB,WFS-2,128 DSBs,\
   dedupe enabled\n\
1029       silver    1  19.9 GB 3.19 GB (16%) 0 B (0%)   NA      \
   6.7 GB (84%)   No                                        4 KB,\
   WFS-2,128 DSBs\n\
1030      cinder1    1  40.8 GB 2.24 GB (5%)  0 B (0%)   NA      \
   38.5 GB (95%)  No                                        4 KB,\
   WFS-2,128 DSBs\n\
1031      cinder2    1  39.8 GB 2.23 GB (6%)  0 B (0%)   NA      \
   37.6 GB (94%)  No                                        4 KB,\
   WFS-2,128 DSBs\n\
1024   fs02-husvm    2  49.8 GB 3.54 GB (7%)  0 B (0%)   NA      \
   46.2 GB (93%)  No                                       32 KB,\
   WFS-2,128 DSBs\n\
1032         test    2  3.97 GB 2.12 GB (53%) 0 B (0%)   NA      \
   1.85 GB (47%)  No                                        4 KB,\
   WFS-2,128 DSBs\n\
1058         huge_FS    7  1.50 TB  Not determined\n\
1053              fs-unmounted    4   108 GB     Not mounted \
   NA  943 MB (18%)  39.2 GB (36%)    No                    4 KB,\
   WFS-2,128 DSBs,dedupe enabled\n\
\n"

HNAS_RESULT6 = "\n\
ID       Label   EVS    Size          Used  Snapshots  Deduped         Avail  \
Thin  ThinSize  ThinAvail               FS Type\n\
----  ---------- ---  ------  ------------  ---------  -------  ------------  \
----  --------  ---------  --------------------\n\
1025  fs01-husvm   1  250 GB  21.4 GB (9%)   0 B (0%)       NA  228 GB (91%)  \
  No                       32 KB,WFS-2,128 DSBs\n\
\n"

HNAS_RESULT7 = "\n\
Export configuration:                       \n\
Export name: /export01-husvm                \n\
Export path: /export01-husvm                \n\
File system label: test_hdp                 \n\
File system size: 250 GB                    \n\
File system free space: 228 GB              \n\
File system state:                          \n\
formatted = Yes                             \n\
mounted = Yes                               \n\
failed = No                                 \n\
thin provisioned = No                       \n\
Access snapshots: Yes                       \n\
Display snapshots: Yes                      \n\
Read Caching: Disabled                      \n\
Disaster recovery setting:                  \n\
Recovered = No                              \n\
Transfer setting = Use file system default  \n\
\n"

HNAS_RESULT8 = "Logical unit creation started at 2014-12-24 00:38:30+00:00."
HNAS_RESULT9 = "Logical unit deleted successfully."
HNAS_RESULT10 = ""
HNAS_RESULT11 = "Logical unit expansion started at 2014-12-24 01:25:03+00:00."

HNAS_RESULT12 = "\n\
Alias               : test_iqn                                       \n\
Globally unique name: iqn.2014-12.10.10.10.10:evstest1.cinder-silver \n\
Comment             :                                                \n\
Secret              : test_secret                                    \n\
Authentication      : Enabled                                        \n\
Logical units       : No logical units.                              \n\
\n"

HNAS_RESULT13 = "Logical unit added successfully."
HNAS_RESULT14 = "Logical unit removed successfully."
HNAS_RESULT15 = "Target created successfully."
HNAS_RESULT16 = ""

HNAS_RESULT17 = "\n\
EVS Type    Label            IP Address          Mask             Port  \n\
----------  ---------------  ------------------  ---------------  ------\n\
evs 1       EVSTest1         172.24.44.20        255.255.255.0    ag1   \n\
evs 2       EVSTest1         10.0.0.20           255.255.255.0    ag1   \n\
\n"

HNAS_RESULT18 = "Version: 11.1.3225.01\n\
Directory: /u/u60/_Eng_Axalon_SMU/OfficialBuilds/fish/angel/3225.01/main/bin/\
x86_64_linux-bart_libc-2.7_release\n\
Date: Feb 22 2013, 04:10:09\n\
\n"

HNAS_RESULT19 = "  ID          Label     Size           Used  Snapshots  \
Deduped          Avail  Thin  ThinSize  ThinAvail              FS Type\n\
----  -------------  -------  -------------  ---------  -------  -------------\
----  --------  ---------  -------------------\n\
1025     fs01-husvm   250 GB  47.1 GB (19%)   0 B (0%)       NA   203 GB (81%)\
  No                       4 KB,WFS-2,128 DSBs\n\
1047  manage_test02  19.9 GB  9.29 GB (47%)   0 B (0%)       NA  10.6 GB (53%)\
  No                       4 KB,WFS-2,128 DSBs\n\
1058         huge_FS    7  1.50 TB  Not determined\n\
1053              fs-unmounted    4   108 GB     Not mounted \
   NA  943 MB (18%)  39.2 GB (36%)    No                 4 KB,\
   WFS-2,128 DSBs,dedupe enabled\n\
\n"

HNAS_RESULT20 = "\n\
Alias               : test_iqn                                       \n\
Globally unique name: iqn.2014-12.10.10.10.10:evstest1.cinder-silver \n\
Comment             :                                                \n\
Secret              :                                                \n\
Authentication      : Enabled                                        \n\
Logical units       : No logical units.                              \n\
\n"

HNAS_RESULT20 = "Target does not exist."

HNAS_RESULT21 = "Target created successfully."

HNAS_RESULT22 = "Failed to establish SSC connection"

HNAS_RESULT23 = "\n\
Alias               : cinder-Gold\n\
Globally unique name: iqn.2015-06.10.10.10.10:evstest1.cinder-gold\n\
Comment             :\n\
Secret              : None\n\
Authentication      : Enabled\n\
Logical units       : No logical units.\n\
Access configuration :\n\
\n\
Alias               : cinder-GoldIsh\n\
Globally unique name: iqn.2015-06.10.10.10.10:evstest1.cinder-goldish\n\
Comment             :\n\
Secret              : None\n\
Authentication      : Enabled\n\
Logical units       : No logical units.\n\
Access configuration :\n\
\n\
Alias               : cinder-default\n\
Globally unique name: iqn.2014-12.10.10.10.10:evstest1.cinder-default\n\
Comment             :\n\
Secret              : pxr6U37LZZJBoMc\n\
Authentication      : Disabled\n\
Logical units       : Logical units       :\n\
\n\
  LUN   Logical Unit\n\
  ----  --------------------------------\n\
  0     volume-8ddd1a54-9daf-4fa5-842...\n\
  1     volume-99da7ae7-1e7f-4d57-8bf...\n\
\n\
Access configuration :\n\
"
HNAS_RESULT24 = "Logical unit modified successfully."

HNAS_RESULT25 = "Current selected file system: HNAS-iSCSI-TEST, number(32)."

HNAS_RESULT26 = "Name   : volume-test \n\
Comment:                              \n\
Path   : /.cinder/volume-test.iscsi   \n\
Size   : 2 GB                         \n\
File System : fs1                     \n\
File System Mounted : YES             \n\
Logical Unit Mounted: No"

HNAS_RESULT27 = "Connection reset"


HNAS_CMDS = {
    ('ssh', '0.0.0.0', 'supervisor', 'supervisor', 'evsfs', 'list'):
        ["%s" % HNAS_RESULT1, ""],
    ('ssh', '0.0.0.0', 'supervisor', 'supervisor', 'cluster-getmac',):
        ["%s" % HNAS_RESULT2, ""],
    ('ssh', '-version',): ["%s" % HNAS_RESULT18, ""],
    ('ssh', '-u', 'supervisor', '-p', 'supervisor', '0.0.0.0', 'ver',):
    ["%s" % HNAS_RESULT3, ""],
    ('ssh', '0.0.0.0', 'supervisor', 'supervisor', 'ver',):
        ["%s" % HNAS_RESULT3, ""],
    ('ssh', '0.0.0.0', 'supervisor', 'supervisor', 'evsipaddr', '-l'):
        ["%s" % HNAS_RESULT4, ""],
    ('ssh', '0.0.0.0', 'supervisor', 'supervisor', 'df', '-a'):
        ["%s" % HNAS_RESULT5, ""],
    ('ssh', '0.0.0.0', 'supervisor', 'supervisor', 'df', '-f', 'test_hdp'):
        ["%s" % HNAS_RESULT6, ""],
    ('ssh', '0.0.0.0', 'supervisor', 'supervisor', 'for-each-evs', '-q',
     'nfs-export', 'list'):
        ["%s" % HNAS_RESULT7, ""],
    ('ssh', '0.0.0.0', 'supervisor', 'supervisor',
     'console-context', '--evs', '1', 'iscsi-lu', 'add', '-e', 'test_name',
     'test_hdp', '/.cinder/test_name.iscsi',
     '1M'):
        ["%s" % HNAS_RESULT8, ""],
    ('ssh', '0.0.0.0', 'supervisor', 'supervisor',
     'console-context', '--evs', '1', 'iscsi-lu', 'del', '-d', '-f',
     'test_lun'):
        ["%s" % HNAS_RESULT9, ""],
    ('ssh', '0.0.0.0', 'supervisor', 'supervisor',
     'console-context', '--evs', '1', 'file-clone-create', '-f', 'fs01-husvm',
     '/.cinder/test_lu.iscsi', 'cloned_lu'):
        ["%s" % HNAS_RESULT10, ""],
    ('ssh', '0.0.0.0', 'supervisor', 'supervisor',
     'console-context', '--evs', '1', 'iscsi-lu', 'expand', 'expanded_lu',
     '1M'):
        ["%s" % HNAS_RESULT11, ""],
    ('ssh', '0.0.0.0', 'supervisor', 'supervisor',
     'console-context', '--evs', '1', 'iscsi-target', 'list', 'test_iqn'):
        ["%s" % HNAS_RESULT12, ""],
    ('ssh', '0.0.0.0', 'supervisor', 'supervisor',
     'console-context', '--evs', '1', 'iscsi-target', 'addlu', 'test_iqn',
     'test_lun', '0'):
        ["%s" % HNAS_RESULT13, ""],
    ('ssh', '0.0.0.0', 'supervisor', 'supervisor',
     'console-context', '--evs', '1', 'iscsi-target', 'dellu', 'test_iqn',
     0):
        ["%s" % HNAS_RESULT14, ""],
    ('ssh', '0.0.0.0', 'supervisor', 'supervisor',
     'console-context', '--evs', '1', 'iscsi-target', 'add', 'myTarget',
     'secret'):
        ["%s" % HNAS_RESULT15, ""],
    ('ssh', '0.0.0.0', 'supervisor', 'supervisor',
     'console-context', '--evs', '1', 'iscsi-target', 'mod', '-s',
     'test_secret', '-a', 'enable', 'test_iqn'): ["%s" % HNAS_RESULT15, ""],
    ('ssh', '0.0.0.0', 'supervisor', 'supervisor',
     'console-context', '--evs', '1', 'iscsi-lu', 'clone', '-e', 'test_lu',
     'test_clone',
     '/.cinder/test_clone.iscsi'):
        ["%s" % HNAS_RESULT16, ""],
    ('ssh', '0.0.0.0', 'supervisor', 'supervisor', 'evsipaddr', '-e', '1'):
        ["%s" % HNAS_RESULT17, ""],
    ('ssh', '0.0.0.0', 'supervisor', 'supervisor',
     'console-context', '--evs', '1', 'iscsi-target', 'list'):
        ["%s" % HNAS_RESULT23, ""],
    ('ssh', '0.0.0.0', 'supervisor', 'supervisor', 'console-context', '--evs',
     '1', 'iscsi-target', 'addlu', 'cinder-default',
     'volume-8ddd1a54-0000-0000-0000', '2'):
        ["%s" % HNAS_RESULT13, ""],
    ('ssh', '0.0.0.0', 'supervisor', 'supervisor', 'console-context', '--evs',
     '1', 'selectfs', 'fs01-husvm'):
        ["%s" % HNAS_RESULT25, ""],
    ('ssh', '0.0.0.0', 'supervisor', 'supervisor', 'console-context', '--evs',
     '1', 'iscsi-lu', 'list', 'test_lun'):
        ["%s" % HNAS_RESULT26, ""],
    ('ssh', '0.0.0.0', 'supervisor', 'supervisor', 'console-context', '--evs',
     '1', 'iscsi-lu', 'mod', '-n', 'vol_test', 'new_vol_test'):
        ["%s" % HNAS_RESULT24, ""]
}

DRV_CONF = {'ssh_enabled': 'True',
            'mgmt_ip0': '0.0.0.0',
            'cluster_admin_ip0': None,
            'ssh_port': '22',
            'ssh_private_key': 'test_key',
            'username': 'supervisor',
            'password': 'supervisor'}

UTILS_EXEC_OUT = ["output: test_cmd", ""]


def m_run_cmd(*args, **kargs):
    return HNAS_CMDS.get(args)


class HDSHNASBendTest(test.TestCase):

    def __init__(self, *args, **kwargs):
        super(HDSHNASBendTest, self).__init__(*args, **kwargs)

    @mock.patch.object(nfs, 'factory_bend')
    def setUp(self, m_factory_bend):
        super(HDSHNASBendTest, self).setUp()
        self.hnas_bend = hnas_backend.HnasBackend(DRV_CONF)

    @mock.patch('six.moves.builtins.open')
    @mock.patch('os.path.isfile', return_value=True)
    @mock.patch('paramiko.RSAKey.from_private_key_file')
    @mock.patch('paramiko.SSHClient')
    @mock.patch.object(putils, 'ssh_execute',
                       return_value=(HNAS_RESULT5, ''))
    @mock.patch.object(utils, 'execute')
    @mock.patch.object(time, 'sleep')
    def test_run_cmd(self, m_sleep, m_utl, m_ssh, m_ssh_cli, m_pvt_key,
                     m_file, m_open):
        save_hkey_file = CONF.ssh_hosts_key_file
        save_spath = CONF.state_path
        CONF.ssh_hosts_key_file = '/var/lib/cinder/ssh_known_hosts'
        CONF.state_path = '/var/lib/cinder'

        # Test main flow
        self.hnas_bend.drv_configs['ssh_enabled'] = 'True'
        out, err = self.hnas_bend.run_cmd('ssh', '0.0.0.0',
                                          'supervisor', 'supervisor',
                                          'df', '-a')
        self.assertIn('fs01-husvm', out)
        self.assertIn('WFS-2,128 DSBs', out)

        # Test exception throwing when not using SSH
        m_utl.side_effect = putils.ProcessExecutionError(stdout='',
                                                         stderr=HNAS_RESULT22,
                                                         exit_code=255)
        self.hnas_bend.drv_configs['ssh_enabled'] = 'False'
        self.assertRaises(exception.HNASConnError, self.hnas_bend.run_cmd,
                          'ssh', '0.0.0.0', 'supervisor', 'supervisor',
                          'df', '-a')

        m_utl.side_effect = putils.ProcessExecutionError(stdout='',
                                                         stderr=HNAS_RESULT27,
                                                         exit_code=255)
        self.hnas_bend.drv_configs['ssh_enabled'] = 'False'
        self.assertRaises(exception.HNASConnError, self.hnas_bend.run_cmd,
                          'ssh', '0.0.0.0', 'supervisor', 'supervisor',
                          'df', '-a')

        # Test exception throwing when using SSH
        m_ssh.side_effect = putils.ProcessExecutionError(stdout='',
                                                         stderr=HNAS_RESULT22,
                                                         exit_code=255)
        self.hnas_bend.drv_configs['ssh_enabled'] = 'True'
        self.assertRaises(exception.HNASConnError, self.hnas_bend.run_cmd,
                          'ssh', '0.0.0.0', 'supervisor', 'supervisor',
                          'df', '-a')

        CONF.state_path = save_spath
        CONF.ssh_hosts_key_file = save_hkey_file

    @mock.patch.object(hnas_backend.HnasBackend, 'run_cmd',
                       side_effect=m_run_cmd)
    @mock.patch.object(utils, 'execute', return_value=UTILS_EXEC_OUT)
    def test_get_version(self, m_cmd, m_exec):
        out = self.hnas_bend.get_version("ssh", "1.0", "0.0.0.0", "supervisor",
                                         "supervisor")
        self.assertIn('11.2.3319.14', out)
        self.assertIn('83-68-96-AA-DA-5D', out)

    @mock.patch.object(hnas_backend.HnasBackend, 'run_cmd',
                       side_effect=m_run_cmd)
    def test_get_version_ssh_cluster(self, m_cmd):
        self.hnas_bend.drv_configs['ssh_enabled'] = 'True'
        self.hnas_bend.drv_configs['cluster_admin_ip0'] = '1.1.1.1'
        out = self.hnas_bend.get_version("ssh", "1.0", "0.0.0.0", "supervisor",
                                         "supervisor")
        self.assertIn('11.2.3319.14', out)
        self.assertIn('83-68-96-AA-DA-5D', out)

    @mock.patch.object(hnas_backend.HnasBackend, 'run_cmd',
                       side_effect=m_run_cmd)
    @mock.patch.object(utils, 'execute', return_value=UTILS_EXEC_OUT)
    def test_get_version_ssh_disable(self, m_cmd, m_exec):
        self.hnas_bend.drv_configs['ssh_enabled'] = 'False'
        out = self.hnas_bend.get_version("ssh", "1.0", "0.0.0.0", "supervisor",
                                         "supervisor")
        self.assertIn('11.2.3319.14', out)
        self.assertIn('83-68-96-AA-DA-5D', out)
        self.assertIn('Utility_version', out)

    @mock.patch.object(hnas_backend.HnasBackend, 'run_cmd',
                       side_effect=m_run_cmd)
    def test_get_iscsi_info(self, m_execute):
        out = self.hnas_bend.get_iscsi_info("ssh", "0.0.0.0", "supervisor",
                                            "supervisor")

        self.assertIn('172.24.44.20', out)
        self.assertIn('172.24.44.21', out)
        self.assertIn('10.0.0.20', out)
        self.assertEqual(4, len(out.split('\n')))

    @mock.patch.object(hnas_backend.HnasBackend, 'run_cmd')
    def test_get_hdp_info(self, m_run_cmd):
        # tests when there is two or more evs
        m_run_cmd.return_value = (HNAS_RESULT5, "")
        out = self.hnas_bend.get_hdp_info("ssh", "0.0.0.0", "supervisor",
                                          "supervisor")

        self.assertEqual(10, len(out.split('\n')))
        self.assertIn('gold', out)
        self.assertIn('silver', out)
        line1 = out.split('\n')[0]
        self.assertEqual(12, len(line1.split()))

        # test when there is only one evs
        m_run_cmd.return_value = (HNAS_RESULT19, "")
        out = self.hnas_bend.get_hdp_info("ssh", "0.0.0.0", "supervisor",
                                          "supervisor")
        self.assertEqual(3, len(out.split('\n')))
        self.assertIn('fs01-husvm', out)
        self.assertIn('manage_test02', out)
        line1 = out.split('\n')[0]
        self.assertEqual(12, len(line1.split()))

    @mock.patch.object(hnas_backend.HnasBackend, 'run_cmd',
                       side_effect=m_run_cmd)
    def test_get_nfs_info(self, m_run_cmd):
        out = self.hnas_bend.get_nfs_info("ssh", "0.0.0.0", "supervisor",
                                          "supervisor")

        self.assertEqual(2, len(out.split('\n')))
        self.assertIn('/export01-husvm', out)
        self.assertIn('172.24.44.20', out)
        self.assertIn('10.0.0.20', out)

    @mock.patch.object(hnas_backend.HnasBackend, 'run_cmd',
                       side_effect=m_run_cmd)
    def test_create_lu(self, m_cmd):
        out = self.hnas_bend.create_lu("ssh", "0.0.0.0", "supervisor",
                                       "supervisor", "test_hdp", "1",
                                       "test_name")

        self.assertIn('successfully created', out)

    @mock.patch.object(hnas_backend.HnasBackend, 'run_cmd',
                       side_effect=m_run_cmd)
    def test_delete_lu(self, m_cmd):
        out = self.hnas_bend.delete_lu("ssh", "0.0.0.0", "supervisor",
                                       "supervisor", "test_hdp", "test_lun")

        self.assertIn('deleted successfully', out)

    @mock.patch.object(hnas_backend.HnasBackend, 'run_cmd',
                       side_effect=m_run_cmd)
    def test_create_dup(self, m_cmd):

        out = self.hnas_bend.create_dup("ssh", "0.0.0.0", "supervisor",
                                        "supervisor", "test_lu", "test_hdp",
                                        "1", "test_clone")

        self.assertIn('successfully created', out)

    @mock.patch.object(hnas_backend.HnasBackend, 'run_cmd',
                       side_effect=m_run_cmd)
    def test_file_clone(self, m_cmd):
        out = self.hnas_bend.file_clone("ssh", "0.0.0.0", "supervisor",
                                        "supervisor", "fs01-husvm",
                                        "/.cinder/test_lu.iscsi", "cloned_lu")

        self.assertIn('LUN cloned_lu HDP', out)

    @mock.patch.object(hnas_backend.HnasBackend, 'run_cmd',
                       side_effect=m_run_cmd)
    def test_extend_vol(self, m_cmd):
        out = self.hnas_bend.extend_vol("ssh", "0.0.0.0", "supervisor",
                                        "supervisor", "test_hdp", "test_lun",
                                        "1", "expanded_lu")

        self.assertIn('successfully extended', out)

    @mock.patch.object(hnas_backend.HnasBackend, 'run_cmd',
                       side_effect=m_run_cmd)
    def test_add_iscsi_conn(self, m_cmd):
        out = self.hnas_bend.add_iscsi_conn("ssh", "0.0.0.0", "supervisor",
                                            "supervisor",
                                            "volume-8ddd1a54-0000-0000-0000",
                                            "test_hdp", "test_port",
                                            "cinder-default", "test_init")

        self.assertIn('successfully paired', out)

    @mock.patch.object(hnas_backend.HnasBackend, 'run_cmd',
                       side_effect=m_run_cmd)
    def test_del_iscsi_conn(self, m_cmd):
        out = self.hnas_bend.del_iscsi_conn("ssh", "0.0.0.0", "supervisor",
                                            "supervisor", "1", "test_iqn", 0)

        self.assertIn('already deleted', out)

    @mock.patch.object(hnas_backend.HnasBackend, 'get_evs', return_value=0)
    @mock.patch.object(hnas_backend.HnasBackend, 'run_cmd')
    def test_get_targetiqn(self, m_cmd, m_get_evs):

        m_cmd.side_effect = [[HNAS_RESULT12, '']]
        out = self.hnas_bend.get_targetiqn("ssh", "0.0.0.0", "supervisor",
                                           "supervisor", "test_iqn",
                                           "test_hdp", "test_secret")

        self.assertEqual('test_iqn', out)

        m_cmd.side_effect = [[HNAS_RESULT20, ''], [HNAS_RESULT21, '']]
        out = self.hnas_bend.get_targetiqn("ssh", "0.0.0.0", "supervisor",
                                           "supervisor", "test_iqn2",
                                           "test_hdp", "test_secret")

        self.assertEqual('test_iqn2', out)

        m_cmd.side_effect = [[HNAS_RESULT20, ''], [HNAS_RESULT21, '']]
        out = self.hnas_bend.get_targetiqn("ssh", "0.0.0.0", "supervisor",
                                           "supervisor", "test_iqn3",
                                           "test_hdp", "")

        self.assertEqual('test_iqn3', out)

    @mock.patch.object(hnas_backend.HnasBackend, 'run_cmd',
                       side_effect=m_run_cmd)
    def test_set_targetsecret(self, m_execute):
        self.hnas_bend.set_targetsecret("ssh", "0.0.0.0", "supervisor",
                                        "supervisor", "test_iqn",
                                        "test_hdp", "test_secret")

    @mock.patch.object(hnas_backend.HnasBackend, 'run_cmd')
    def test_get_targetsecret(self, m_run_cmd):
        # test when target has secret
        m_run_cmd.return_value = (HNAS_RESULT12, "")
        out = self.hnas_bend.get_targetsecret("ssh", "0.0.0.0", "supervisor",
                                              "supervisor", "test_iqn",
                                              "test_hdp")

        self.assertEqual('test_secret', out)

        # test when target don't have secret
        m_run_cmd.return_value = (HNAS_RESULT20, "")
        out = self.hnas_bend.get_targetsecret("ssh", "0.0.0.0", "supervisor",
                                              "supervisor", "test_iqn",
                                              "test_hdp")
        self.assertEqual('', out)

    @mock.patch.object(hnas_backend.HnasBackend, 'run_cmd')
    def test_get_targets(self, m_run_cmd):
        # Test normal behaviour
        m_run_cmd.return_value = (HNAS_RESULT23, "")
        tgt_list = self.hnas_bend._get_targets("ssh", "0.0.0.0", "supervisor",
                                               "supervisor", 1)
        self.assertEqual(3, len(tgt_list))
        self.assertEqual(2, len(tgt_list[2]['luns']))

        # Test calling with parameter
        tgt_list = self.hnas_bend._get_targets("ssh", "0.0.0.0", "supervisor",
                                               "supervisor", 1,
                                               'cinder-default')
        self.assertEqual(1, len(tgt_list))
        self.assertEqual(2, len(tgt_list[0]['luns']))

        # Test error in BE command
        m_run_cmd.side_effect = putils.ProcessExecutionError
        tgt_list = self.hnas_bend._get_targets("ssh", "0.0.0.0", "supervisor",
                                               "supervisor", 1)
        self.assertEqual(0, len(tgt_list))

    @mock.patch.object(hnas_backend.HnasBackend,
                       'run_cmd', side_effect=m_run_cmd)
    def test_check_targets(self, m_run_cmd):
        result, tgt = self.hnas_bend.check_target("ssh", "0.0.0.0",
                                                  "supervisor",
                                                  "supervisor", "test_hdp",
                                                  "cinder-default")
        self.assertTrue(result)
        self.assertEqual('cinder-default', tgt['alias'])

        result, tgt = self.hnas_bend.check_target("ssh", "0.0.0.0",
                                                  "supervisor",
                                                  "supervisor", "test_hdp",
                                                  "cinder-no-target")
        self.assertFalse(result)
        self.assertIsNone(tgt)

    @mock.patch.object(hnas_backend.HnasBackend,
                       'run_cmd', side_effect=m_run_cmd)
    def test_check_lu(self, m_run_cmd):
        ret = self.hnas_bend.check_lu("ssh", "0.0.0.0", "supervisor",
                                      "supervisor",
                                      "volume-8ddd1a54-9daf-4fa5-842",
                                      "test_hdp")
        result, lunid, tgt = ret
        self.assertTrue(result)
        self.assertEqual('0', lunid)

        ret = self.hnas_bend.check_lu("ssh", "0.0.0.0", "supervisor",
                                      "supervisor",
                                      "volume-8ddd1a54-0000-0000-000",
                                      "test_hdp")
        result, lunid, tgt = ret
        self.assertFalse(result)

    @mock.patch.object(hnas_backend.HnasBackend, 'get_evs', return_value=1)
    @mock.patch.object(hnas_backend.HnasBackend, 'run_cmd',
                       return_value = (HNAS_RESULT26, ""))
    def test_get_existing_lu_info(self, m_run_cmd, m_get_evs):

        out = self.hnas_bend.get_existing_lu_info("ssh", "0.0.0.0",
                                                  "supervisor",
                                                  "supervisor", "fs01-husvm",
                                                  "test_lun")

        m_get_evs.assert_called_once_with('ssh', '0.0.0.0', 'supervisor',
                                          'supervisor', 'fs01-husvm')
        m_run_cmd.assert_called_once_with('ssh', '0.0.0.0', 'supervisor',
                                          'supervisor', 'console-context',
                                          '--evs', 1, 'iscsi-lu', 'list',
                                          'test_lun')

        self.assertEqual(HNAS_RESULT26, out)

    @mock.patch.object(hnas_backend.HnasBackend, 'get_evs', return_value=1)
    @mock.patch.object(hnas_backend.HnasBackend, 'run_cmd',
                       return_value=(HNAS_RESULT24, ""))
    def test_rename_existing_lu(self, m_run_cmd, m_get_evs):

        out = self.hnas_bend.rename_existing_lu("ssh", "0.0.0.0",
                                                "supervisor",
                                                "supervisor", "fs01-husvm",
                                                "vol_test",
                                                "new_vol_test")

        m_get_evs.assert_called_once_with('ssh', '0.0.0.0', 'supervisor',
                                          'supervisor', 'fs01-husvm')
        m_run_cmd.assert_called_once_with('ssh', '0.0.0.0', 'supervisor',
                                          'supervisor', 'console-context',
                                          '--evs', 1, 'iscsi-lu', 'mod',
                                          '-n', 'vol_test', 'new_vol_test')

        self.assertEqual(HNAS_RESULT24, out)



# Copyright 2010 OpenStack Foundation
# Copyright 2013 NTT corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Implementation of an image service that uses Glance as the backend"""


from __future__ import absolute_import

import copy
import itertools
import random
import shutil
import sys
import time

import glanceclient.exc
from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import timeutils
import six
from six.moves import range
from six.moves import urllib

from cinder import exception
from cinder.i18n import _LE, _LW


glance_opts = [
    cfg.ListOpt('allowed_direct_url_schemes',
                default=[],
                help='A list of url schemes that can be downloaded directly '
                     'via the direct_url.  Currently supported schemes: '
                     '[file].'),
]
glance_core_properties_opts = [
    cfg.ListOpt('glance_core_properties',
                default=['checksum', 'container_format',
                         'disk_format', 'image_name', 'image_id',
                         'min_disk', 'min_ram', 'name', 'size'],
                help='Default core properties of image')
]
CONF = cfg.CONF
CONF.register_opts(glance_opts)
CONF.register_opts(glance_core_properties_opts)
CONF.import_opt('glance_api_version', 'cinder.common.config')

LOG = logging.getLogger(__name__)


def _parse_image_ref(image_href):
    """Parse an image href into composite parts.

    :param image_href: href of an image
    :returns: a tuple of the form (image_id, netloc, use_ssl)
    :raises ValueError

    """
    url = urllib.parse.urlparse(image_href)
    netloc = url.netloc
    image_id = url.path.split('/')[-1]
    use_ssl = (url.scheme == 'https')
    return (image_id, netloc, use_ssl)


def _create_glance_client(context, netloc, use_ssl, version=None):
    """Instantiate a new glanceclient.Client object."""
    if version is None:
        version = CONF.glance_api_version
    params = {}
    if use_ssl:
        scheme = 'https'
        # https specific params
        params['insecure'] = CONF.glance_api_insecure
        params['ssl_compression'] = CONF.glance_api_ssl_compression
        params['cacert'] = CONF.glance_ca_certificates_file
    else:
        scheme = 'http'
    if CONF.auth_strategy == 'keystone':
        params['token'] = context.auth_token
    if CONF.glance_request_timeout is not None:
        params['timeout'] = CONF.glance_request_timeout
    endpoint = '%s://%s' % (scheme, netloc)
    return glanceclient.Client(str(version), endpoint, **params)


def get_api_servers():
    """Return Iterable over shuffled api servers.

    Shuffle a list of CONF.glance_api_servers and return an iterator
    that will cycle through the list, looping around to the beginning
    if necessary.
    """
    api_servers = []
    for api_server in CONF.glance_api_servers:
        if '//' not in api_server:
            api_server = 'http://' + api_server
        url = urllib.parse.urlparse(api_server)
        netloc = url.netloc
        use_ssl = (url.scheme == 'https')
        api_servers.append((netloc, use_ssl))
    random.shuffle(api_servers)
    return itertools.cycle(api_servers)


class GlanceClientWrapper(object):
    """Glance client wrapper class that implements retries."""

    def __init__(self, context=None, netloc=None, use_ssl=False,
                 version=None):
        if netloc is not None:
            self.client = self._create_static_client(context,
                                                     netloc,
                                                     use_ssl, version)
        else:
            self.client = None
        self.api_servers = None
        self.version = version

        if CONF.glance_num_retries < 0:
            LOG.warning(_LW(
                "glance_num_retries shouldn't be a negative value. "
                "The number of retries will be set to 0 until this is"
                "corrected in the cinder.conf."))
            CONF.set_override('glance_num_retries', 0)

    def _create_static_client(self, context, netloc, use_ssl, version):
        """Create a client that we'll use for every call."""
        self.netloc = netloc
        self.use_ssl = use_ssl
        self.version = version
        return _create_glance_client(context,
                                     self.netloc,
                                     self.use_ssl, self.version)

    def _create_onetime_client(self, context, version):
        """Create a client that will be used for one call."""
        if self.api_servers is None:
            self.api_servers = get_api_servers()
        self.netloc, self.use_ssl = next(self.api_servers)
        return _create_glance_client(context,
                                     self.netloc,
                                     self.use_ssl, version)

    def call(self, context, method, *args, **kwargs):
        """Call a glance client method.

        If we get a connection error,
        retry the request according to CONF.glance_num_retries.
        """
        version = kwargs.pop('version', self.version)

        retry_excs = (glanceclient.exc.ServiceUnavailable,
                      glanceclient.exc.InvalidEndpoint,
                      glanceclient.exc.CommunicationError)
        num_attempts = 1 + CONF.glance_num_retries

        for attempt in range(1, num_attempts + 1):
            client = self.client or self._create_onetime_client(context,
                                                                version)
            try:
                controller = getattr(client,
                                     kwargs.pop('controller', 'images'))
                return getattr(controller, method)(*args, **kwargs)
            except retry_excs as e:
                netloc = self.netloc
                extra = "retrying"
                error_msg = _LE("Error contacting glance server "
                                "'%(netloc)s' for '%(method)s', "
                                "%(extra)s.")
                if attempt == num_attempts:
                    extra = 'done trying'
                    LOG.exception(error_msg, {'netloc': netloc,
                                              'method': method,
                                              'extra': extra})
                    raise exception.GlanceConnectionFailed(reason=e)

                LOG.exception(error_msg, {'netloc': netloc,
                                          'method': method,
                                          'extra': extra})
                time.sleep(1)


class GlanceImageService(object):
    """Provides storage and retrieval of disk image objects within Glance."""

    def __init__(self, client=None):
        self._client = client or GlanceClientWrapper()
        self._image_schema = None
        self.temp_images = None

    def detail(self, context, **kwargs):
        """Calls out to Glance for a list of detailed image information."""
        params = self._extract_query_params(kwargs)
        try:
            images = self._client.call(context, 'list', **params)
        except Exception:
            _reraise_translated_exception()

        _images = []
        for image in images:
            if self._is_image_available(context, image):
                _images.append(self._translate_from_glance(context, image))

        return _images

    def _extract_query_params(self, params):
        _params = {}
        accepted_params = ('filters', 'marker', 'limit',
                           'sort_key', 'sort_dir')
        for param in accepted_params:
            if param in params:
                _params[param] = params.get(param)

        # ensure filters is a dict
        _params.setdefault('filters', {})
        # NOTE(vish): don't filter out private images
        _params['filters'].setdefault('is_public', 'none')

        return _params

    def show(self, context, image_id):
        """Returns a dict with image data for the given opaque image id."""
        try:
            image = self._client.call(context, 'get', image_id)
        except Exception:
            _reraise_translated_image_exception(image_id)

        if not self._is_image_available(context, image):
            raise exception.ImageNotFound(image_id=image_id)

        base_image_meta = self._translate_from_glance(context, image)
        return base_image_meta

    def get_location(self, context, image_id):
        """Get backend storage location url.

        Returns a tuple containing the direct url and locations representing
        the backend storage location, or (None, None) if these attributes are
        not shown by Glance.
        """
        if CONF.glance_api_version == 1:
            # image location not available in v1
            return (None, None)
        try:
            # direct_url is returned by v2 api
            client = GlanceClientWrapper(version=2)
            image_meta = client.call(context, 'get', image_id)
        except Exception:
            _reraise_translated_image_exception(image_id)

        if not self._is_image_available(context, image_meta):
            raise exception.ImageNotFound(image_id=image_id)

        # some glance stores like nfs only meta data
        # is stored and returned as locations.
        # so composite of two needs to be returned.
        return (getattr(image_meta, 'direct_url', None),
                getattr(image_meta, 'locations', None))

    def add_location(self, context, image_id, url, metadata):
        """Add a backend location url to an image.

        Returns a dict containing image metadata on success.
        """
        if CONF.glance_api_version != 2:
            raise exception.Invalid("Image API version 2 is disabled.")
        client = GlanceClientWrapper(version=2)
        try:
            return client.call(context, 'add_location',
                               image_id, url, metadata)
        except Exception:
            _reraise_translated_image_exception(image_id)

    def delete_locations(self, context, image_id, url_set):
        """Delete backend location urls from an image."""
        if CONF.glance_api_version != 2:
            raise exception.Invalid("Image API version 2 is disabled.")
        client = GlanceClientWrapper(version=2)
        try:
            return client.call(context, 'delete_locations', image_id, url_set)
        except Exception:
            _reraise_translated_image_exception(image_id)

    def download(self, context, image_id, data=None):
        """Calls out to Glance for data and writes data."""
        if data and 'file' in CONF.allowed_direct_url_schemes:
            direct_url, locations = self.get_location(context, image_id)
            urls = [direct_url] + [loc.get('url') for loc in locations or []]
            for url in urls:
                if url is None:
                    continue
                parsed_url = urllib.parse.urlparse(url)
                if parsed_url.scheme == "file":
                    # a system call to cp could have significant performance
                    # advantages, however we do not have the path to files at
                    # this point in the abstraction.
                    with open(parsed_url.path, "r") as f:
                        shutil.copyfileobj(f, data)
                    return

        try:
            image_chunks = self._client.call(context, 'data', image_id)
        except Exception:
            _reraise_translated_image_exception(image_id)

        if not data:
            return image_chunks
        else:
            for chunk in image_chunks:
                data.write(chunk)

    def create(self, context, image_meta, data=None):
        """Store the image data and return the new image object."""
        sent_service_image_meta = self._translate_to_glance(image_meta)

        if data:
            sent_service_image_meta['data'] = data

        recv_service_image_meta = self._client.call(context, 'create',
                                                    **sent_service_image_meta)

        return self._translate_from_glance(context, recv_service_image_meta)

    def update(self, context, image_id,
               image_meta, data=None, purge_props=True):
        """Modify the given image with the new data."""
        image_meta = self._translate_to_glance(image_meta)
        # NOTE(dosaboy): see comment in bug 1210467
        if CONF.glance_api_version == 1:
            image_meta['purge_props'] = purge_props
        # NOTE(bcwaldon): id is not an editable field, but it is likely to be
        # passed in by calling code. Let's be nice and ignore it.
        image_meta.pop('id', None)
        if data:
            image_meta['data'] = data
        try:
            # NOTE(dosaboy): the v2 api separates update from upload
            if data and CONF.glance_api_version > 1:
                self._client.call(context, 'upload', image_id, data)
                image_meta = self._client.call(context, 'get', image_id)
            else:
                image_meta = self._client.call(context, 'update', image_id,
                                               **image_meta)
        except Exception:
            _reraise_translated_image_exception(image_id)
        else:
            return self._translate_from_glance(context, image_meta)

    def delete(self, context, image_id):
        """Delete the given image.

        :raises: ImageNotFound if the image does not exist.
        :raises: NotAuthorized if the user is not an owner.

        """
        try:
            self._client.call(context, 'delete', image_id)
        except glanceclient.exc.NotFound:
            raise exception.ImageNotFound(image_id=image_id)
        return True

    def _translate_from_glance(self, context, image):
        """Get image metadata from glance image.

        Extract metadata from image and convert it's properties
        to type cinder expected.

        :param image: glance image object
        :return: image metadata dictionary
        """
        if CONF.glance_api_version == 2:
            if self._image_schema is None:
                self._image_schema = self._client.call(context, 'get',
                                                       controller='schemas',
                                                       schema_name='image',
                                                       version=2)
            # NOTE(aarefiev): get base image property, store image 'schema'
            #                 is redundant, so ignore it.
            image_meta = {key: getattr(image, key)
                          for key in image.keys()
                          if self._image_schema.is_base_property(key) is True
                          and key != 'schema'}

            # NOTE(aarefiev): nova is expected that all image properties
            # (custom or defined in schema-image.json) stores in
            # 'properties' key.
            image_meta['properties'] = {
                key: getattr(image, key) for key in image.keys()
                if self._image_schema.is_base_property(key) is False}
        else:
            image_meta = _extract_attributes(image)

        image_meta = _convert_timestamps_to_datetimes(image_meta)
        image_meta = _convert_from_string(image_meta)
        return image_meta

    @staticmethod
    def _translate_to_glance(image_meta):
        image_meta = _convert_to_string(image_meta)
        image_meta = _remove_read_only(image_meta)

        # NOTE(tsekiyama): From the Image API v2, custom properties must
        # be stored in image_meta directly, instead of the 'properties' key.
        if CONF.glance_api_version >= 2:
            properties = image_meta.get('properties')
            if properties:
                image_meta.update(properties)
                del image_meta['properties']

        return image_meta

    @staticmethod
    def _is_image_available(context, image):
        """Check image availability.

        This check is needed in case Nova and Glance are deployed
        without authentication turned on.
        """
        # The presence of an auth token implies this is an authenticated
        # request and we need not handle the noauth use-case.
        if hasattr(context, 'auth_token') and context.auth_token:
            return True

        if image.is_public or context.is_admin:
            return True

        properties = image.properties

        if context.project_id and ('owner_id' in properties):
            return str(properties['owner_id']) == str(context.project_id)

        if context.project_id and ('project_id' in properties):
            return str(properties['project_id']) == str(context.project_id)

        try:
            user_id = properties['user_id']
        except KeyError:
            return False

        return str(user_id) == str(context.user_id)


def _convert_timestamps_to_datetimes(image_meta):
    """Returns image with timestamp fields converted to datetime objects."""
    for attr in ['created_at', 'updated_at', 'deleted_at']:
        if image_meta.get(attr):
            image_meta[attr] = timeutils.parse_isotime(image_meta[attr])
    return image_meta


# NOTE(bcwaldon): used to store non-string data in glance metadata
def _json_loads(properties, attr):
    prop = properties[attr]
    if isinstance(prop, six.string_types):
        properties[attr] = jsonutils.loads(prop)


def _json_dumps(properties, attr):
    prop = properties[attr]
    if not isinstance(prop, six.string_types):
        properties[attr] = jsonutils.dumps(prop)


_CONVERT_PROPS = ('block_device_mapping', 'mappings')


def _convert(method, metadata):
    metadata = copy.deepcopy(metadata)
    properties = metadata.get('properties')
    if properties:
        for attr in _CONVERT_PROPS:
            if attr in properties:
                method(properties, attr)

    return metadata


def _convert_from_string(metadata):
    return _convert(_json_loads, metadata)


def _convert_to_string(metadata):
    return _convert(_json_dumps, metadata)


def _extract_attributes(image):
    # NOTE(hdd): If a key is not found, base.Resource.__getattr__() may perform
    # a get(), resulting in a useless request back to glance. This list is
    # therefore sorted, with dependent attributes as the end
    # 'deleted_at' depends on 'deleted'
    # 'checksum' depends on 'status' == 'active'
    IMAGE_ATTRIBUTES = ['size', 'disk_format', 'owner',
                        'container_format', 'status', 'id',
                        'name', 'created_at', 'updated_at',
                        'deleted', 'deleted_at', 'checksum',
                        'min_disk', 'min_ram', 'is_public']
    output = {}

    for attr in IMAGE_ATTRIBUTES:
        if attr == 'deleted_at' and not output['deleted']:
            output[attr] = None
        elif attr == 'checksum' and output['status'] != 'active':
            output[attr] = None
        else:
            output[attr] = getattr(image, attr, None)

    output['properties'] = getattr(image, 'properties', {})

    # NOTE(jbernard): Update image properties for API version 2.  For UEC
    # images stored in glance, the necessary boot information is stored in the
    # properties dict in version 1 so there is nothing more to do.  However, in
    # version 2 these are standalone fields in the GET response.  This bit of
    # code moves them back into the properties dict as the caller expects, thus
    # producing a volume with correct metadata for booting.
    for attr in ('kernel_id', 'ramdisk_id'):
        value = getattr(image, attr, None)
        if value:
            output['properties'][attr] = value

    return output


def _remove_read_only(image_meta):
    IMAGE_ATTRIBUTES = ['status', 'updated_at', 'created_at', 'deleted_at']
    output = copy.deepcopy(image_meta)
    for attr in IMAGE_ATTRIBUTES:
        if attr in output:
            del output[attr]
    return output


def _reraise_translated_image_exception(image_id):
    """Transform the exception for the image but keep its traceback intact."""
    _exc_type, exc_value, exc_trace = sys.exc_info()
    new_exc = _translate_image_exception(image_id, exc_value)
    six.reraise(type(new_exc), new_exc, exc_trace)


def _reraise_translated_exception():
    """Transform the exception but keep its traceback intact."""
    _exc_type, exc_value, exc_trace = sys.exc_info()
    new_exc = _translate_plain_exception(exc_value)
    six.reraise(type(new_exc), new_exc, exc_trace)


def _translate_image_exception(image_id, exc_value):
    if isinstance(exc_value, (glanceclient.exc.Forbidden,
                              glanceclient.exc.Unauthorized)):
        return exception.ImageNotAuthorized(image_id=image_id)
    if isinstance(exc_value, glanceclient.exc.NotFound):
        return exception.ImageNotFound(image_id=image_id)
    if isinstance(exc_value, glanceclient.exc.BadRequest):
        return exception.Invalid(exc_value)
    return exc_value


def _translate_plain_exception(exc_value):
    if isinstance(exc_value, (glanceclient.exc.Forbidden,
                              glanceclient.exc.Unauthorized)):
        return exception.NotAuthorized(exc_value)
    if isinstance(exc_value, glanceclient.exc.NotFound):
        return exception.NotFound(exc_value)
    if isinstance(exc_value, glanceclient.exc.BadRequest):
        return exception.Invalid(exc_value)
    return exc_value


def get_remote_image_service(context, image_href):
    """Create an image_service and parse the id from the given image_href.

    The image_href param can be an href of the form
    'http://example.com:9292/v1/images/b8b2c6f7-7345-4e2f-afa2-eedaba9cbbe3',
    or just an id such as 'b8b2c6f7-7345-4e2f-afa2-eedaba9cbbe3'. If the
    image_href is a standalone id, then the default image service is returned.

    :param image_href: href that describes the location of an image
    :returns: a tuple of the form (image_service, image_id)

    """
    # NOTE(bcwaldon): If image_href doesn't look like a URI, assume its a
    # standalone image ID
    if '/' not in str(image_href):
        image_service = get_default_image_service()
        return image_service, image_href

    try:
        (image_id, glance_netloc, use_ssl) = _parse_image_ref(image_href)
        glance_client = GlanceClientWrapper(context=context,
                                            netloc=glance_netloc,
                                            use_ssl=use_ssl)
    except ValueError:
        raise exception.InvalidImageRef(image_href=image_href)

    image_service = GlanceImageService(client=glance_client)
    return image_service, image_id


def get_default_image_service():
    return GlanceImageService()



# Copyright (c) 2015 HGST Inc
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import mock

from oslo_concurrency import processutils

from cinder import context
from cinder import exception
from cinder import test
from cinder.volume import configuration as conf
from cinder.volume.drivers.hgst import HGSTDriver
from cinder.volume import volume_types


class HGSTTestCase(test.TestCase):

    # Need to mock these since we use them on driver creation
    @mock.patch('pwd.getpwnam', return_value=1)
    @mock.patch('grp.getgrnam', return_value=1)
    @mock.patch('socket.gethostbyname', return_value='123.123.123.123')
    def setUp(self, mock_ghn, mock_grnam, mock_pwnam):
        """Set up UUT and all the flags required for later fake_executes."""
        super(HGSTTestCase, self).setUp()
        self.stubs.Set(processutils, 'execute', self._fake_execute)
        self._fail_vgc_cluster = False
        self._fail_ip = False
        self._fail_network_list = False
        self._fail_domain_list = False
        self._empty_domain_list = False
        self._fail_host_storage = False
        self._fail_space_list = False
        self._fail_space_delete = False
        self._fail_set_apphosts = False
        self._fail_extend = False
        self._request_cancel = False
        self._return_blocked = 0
        self.configuration = mock.Mock(spec=conf.Configuration)
        self.configuration.safe_get = self._fake_safe_get
        self._reset_configuration()
        self.driver = HGSTDriver(configuration=self.configuration,
                                 execute=self._fake_execute)

    def _fake_safe_get(self, value):
        """Don't throw exception on missing parameters, return None."""
        try:
            val = getattr(self.configuration, value)
        except AttributeError:
            val = None
        return val

    def _reset_configuration(self):
        """Set safe and sane values for config params."""
        self.configuration.num_volume_device_scan_tries = 1
        self.configuration.volume_dd_blocksize = '1M'
        self.configuration.volume_backend_name = 'hgst-1'
        self.configuration.hgst_storage_servers = 'stor1:gbd0,stor2:gbd0'
        self.configuration.hgst_net = 'net1'
        self.configuration.hgst_redundancy = '0'
        self.configuration.hgst_space_user = 'kane'
        self.configuration.hgst_space_group = 'xanadu'
        self.configuration.hgst_space_mode = '0777'

    def _parse_space_create(self, *cmd):
        """Eats a vgc-cluster space-create command line to a dict."""
        self.created = {'storageserver': ''}
        cmd = list(*cmd)
        while cmd:
            param = cmd.pop(0)
            if param == "-n":
                self.created['name'] = cmd.pop(0)
            elif param == "-N":
                self.created['net'] = cmd.pop(0)
            elif param == "-s":
                self.created['size'] = cmd.pop(0)
            elif param == "--redundancy":
                self.created['redundancy'] = cmd.pop(0)
            elif param == "--user":
                self.created['user'] = cmd.pop(0)
            elif param == "--user":
                self.created['user'] = cmd.pop(0)
            elif param == "--group":
                self.created['group'] = cmd.pop(0)
            elif param == "--mode":
                self.created['mode'] = cmd.pop(0)
            elif param == "-S":
                self.created['storageserver'] += cmd.pop(0) + ","
            else:
                pass

    def _parse_space_extend(self, *cmd):
        """Eats a vgc-cluster space-extend commandline to a dict."""
        self.extended = {'storageserver': ''}
        cmd = list(*cmd)
        while cmd:
            param = cmd.pop(0)
            if param == "-n":
                self.extended['name'] = cmd.pop(0)
            elif param == "-s":
                self.extended['size'] = cmd.pop(0)
            elif param == "-S":
                self.extended['storageserver'] += cmd.pop(0) + ","
            else:
                pass
        if self._fail_extend:
            raise processutils.ProcessExecutionError(exit_code=1)
        else:
            return '', ''

    def _parse_space_delete(self, *cmd):
        """Eats a vgc-cluster space-delete commandline to a dict."""
        self.deleted = {}
        cmd = list(*cmd)
        while cmd:
            param = cmd.pop(0)
            if param == "-n":
                self.deleted['name'] = cmd.pop(0)
            else:
                pass
        if self._fail_space_delete:
            raise processutils.ProcessExecutionError(exit_code=1)
        else:
            return '', ''

    def _parse_space_list(self, *cmd):
        """Eats a vgc-cluster space-list commandline to a dict."""
        json = False
        nameOnly = False
        cmd = list(*cmd)
        while cmd:
            param = cmd.pop(0)
            if param == "--json":
                json = True
            elif param == "--name-only":
                nameOnly = True
            elif param == "-n":
                pass  # Don't use the name here...
            else:
                pass
        if self._fail_space_list:
            raise processutils.ProcessExecutionError(exit_code=1)
        elif nameOnly:
            return "space1\nspace2\nvolume1\n", ''
        elif json:
            return HGST_SPACE_JSON, ''
        else:
            return '', ''

    def _parse_network_list(self, *cmd):
        """Eat a network-list command and return error or results."""
        if self._fail_network_list:
            raise processutils.ProcessExecutionError(exit_code=1)
        else:
            return NETWORK_LIST, ''

    def _parse_domain_list(self, *cmd):
        """Eat a domain-list command and return error, empty, or results."""
        if self._fail_domain_list:
            raise processutils.ProcessExecutionError(exit_code=1)
        elif self._empty_domain_list:
            return '', ''
        else:
            return "thisserver\nthatserver\nanotherserver\n", ''

    def _fake_execute(self, *cmd, **kwargs):
        """Sudo hook to catch commands to allow running on all hosts."""
        cmdlist = list(cmd)
        exe = cmdlist.pop(0)
        if exe == 'vgc-cluster':
            exe = cmdlist.pop(0)
            if exe == "request-cancel":
                self._request_cancel = True
                if self._return_blocked > 0:
                    return 'Request cancelled', ''
                else:
                    raise processutils.ProcessExecutionError(exit_code=1)
            elif self._fail_vgc_cluster:
                raise processutils.ProcessExecutionError(exit_code=1)
            elif exe == "--version":
                return "HGST Solutions V2.5.0.0.x.x.x.x.x", ''
            elif exe == "space-list":
                return self._parse_space_list(cmdlist)
            elif exe == "space-create":
                self._parse_space_create(cmdlist)
                if self._return_blocked > 0:
                    self._return_blocked = self._return_blocked - 1
                    out = "VGC_CREATE_000002\nBLOCKED\n"
                    raise processutils.ProcessExecutionError(stdout=out,
                                                             exit_code=1)
                return '', ''
            elif exe == "space-delete":
                return self._parse_space_delete(cmdlist)
            elif exe == "space-extend":
                return self._parse_space_extend(cmdlist)
            elif exe == "host-storage":
                if self._fail_host_storage:
                    raise processutils.ProcessExecutionError(exit_code=1)
                return HGST_HOST_STORAGE, ''
            elif exe == "domain-list":
                return self._parse_domain_list()
            elif exe == "network-list":
                return self._parse_network_list()
            elif exe == "space-set-apphosts":
                if self._fail_set_apphosts:
                    raise processutils.ProcessExecutionError(exit_code=1)
                return '', ''
            else:
                raise NotImplementedError
        elif exe == 'ip':
            if self._fail_ip:
                raise processutils.ProcessExecutionError(exit_code=1)
            else:
                return IP_OUTPUT, ''
        elif exe == 'dd':
            self.dd_count = -1
            for p in cmdlist:
                if 'count=' in p:
                    self.dd_count = int(p[6:])
            return DD_OUTPUT, ''
        else:
            return '', ''

    @mock.patch('pwd.getpwnam', return_value=1)
    @mock.patch('grp.getgrnam', return_value=1)
    @mock.patch('socket.gethostbyname', return_value='123.123.123.123')
    def test_vgc_cluster_not_present(self, mock_ghn, mock_grnam, mock_pwnam):
        """Test exception when vgc-cluster returns an error."""
        # Should pass
        self._fail_vgc_cluster = False
        self.driver.check_for_setup_error()
        # Should throw exception
        self._fail_vgc_cluster = True
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.check_for_setup_error)

    @mock.patch('pwd.getpwnam', return_value=1)
    @mock.patch('grp.getgrnam', return_value=1)
    @mock.patch('socket.gethostbyname', return_value='123.123.123.123')
    def test_parameter_redundancy_invalid(self, mock_ghn, mock_grnam,
                                          mock_pwnam):
        """Test when hgst_redundancy config parameter not 0 or 1."""
        # Should pass
        self.driver.check_for_setup_error()
        # Should throw exceptions
        self.configuration.hgst_redundancy = ''
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.check_for_setup_error)
        self.configuration.hgst_redundancy = 'Fred'
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.check_for_setup_error)

    @mock.patch('pwd.getpwnam', return_value=1)
    @mock.patch('grp.getgrnam', return_value=1)
    @mock.patch('socket.gethostbyname', return_value='123.123.123.123')
    def test_parameter_user_invalid(self, mock_ghn, mock_grnam, mock_pwnam):
        """Test exception when hgst_space_user doesn't map to UNIX user."""
        # Should pass
        self.driver.check_for_setup_error()
        # Should throw exceptions
        mock_pwnam.side_effect = KeyError()
        self.configuration.hgst_space_user = ''
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.check_for_setup_error)
        self.configuration.hgst_space_user = 'Fred!`'
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.check_for_setup_error)

    @mock.patch('pwd.getpwnam', return_value=1)
    @mock.patch('grp.getgrnam', return_value=1)
    @mock.patch('socket.gethostbyname', return_value='123.123.123.123')
    def test_parameter_group_invalid(self, mock_ghn, mock_grnam, mock_pwnam):
        """Test exception when hgst_space_group doesn't map to UNIX group."""
        # Should pass
        self.driver.check_for_setup_error()
        # Should throw exceptions
        mock_grnam.side_effect = KeyError()
        self.configuration.hgst_space_group = ''
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.check_for_setup_error)
        self.configuration.hgst_space_group = 'Fred!`'
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.check_for_setup_error)

    @mock.patch('pwd.getpwnam', return_value=1)
    @mock.patch('grp.getgrnam', return_value=1)
    @mock.patch('socket.gethostbyname', return_value='123.123.123.123')
    def test_parameter_mode_invalid(self, mock_ghn, mock_grnam, mock_pwnam):
        """Test exception when mode for created spaces isn't proper format."""
        # Should pass
        self.driver.check_for_setup_error()
        # Should throw exceptions
        self.configuration.hgst_space_mode = ''
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.check_for_setup_error)
        self.configuration.hgst_space_mode = 'Fred'
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.check_for_setup_error)

    @mock.patch('pwd.getpwnam', return_value=1)
    @mock.patch('grp.getgrnam', return_value=1)
    @mock.patch('socket.gethostbyname', return_value='123.123.123.123')
    def test_parameter_net_invalid(self, mock_ghn, mock_grnam, mock_pwnam):
        """Test exception when hgst_net not in the domain."""
        # Should pass
        self.driver.check_for_setup_error()
        # Should throw exceptions
        self._fail_network_list = True
        self.configuration.hgst_net = 'Fred'
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.check_for_setup_error)
        self._fail_network_list = False

    @mock.patch('pwd.getpwnam', return_value=1)
    @mock.patch('grp.getgrnam', return_value=1)
    @mock.patch('socket.gethostbyname', return_value='123.123.123.123')
    def test_ip_addr_fails(self, mock_ghn, mock_grnam, mock_pwnam):
        """Test exception when IP ADDR command fails."""
        # Should pass
        self.driver.check_for_setup_error()
        # Throw exception, need to clear internal cached host in driver
        self._fail_ip = True
        self.driver._vgc_host = None
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.check_for_setup_error)

    @mock.patch('pwd.getpwnam', return_value=1)
    @mock.patch('grp.getgrnam', return_value=1)
    @mock.patch('socket.gethostbyname', return_value='123.123.123.123')
    def test_domain_list_fails(self, mock_ghn, mock_grnam, mock_pwnam):
        """Test exception when domain-list fails for the domain."""
        # Should pass
        self.driver.check_for_setup_error()
        # Throw exception, need to clear internal cached host in driver
        self._fail_domain_list = True
        self.driver._vgc_host = None
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.check_for_setup_error)

    @mock.patch('pwd.getpwnam', return_value=1)
    @mock.patch('grp.getgrnam', return_value=1)
    @mock.patch('socket.gethostbyname', return_value='123.123.123.123')
    def test_not_in_domain(self, mock_ghn, mock_grnam, mock_pwnam):
        """Test exception when Cinder host not domain member."""
        # Should pass
        self.driver.check_for_setup_error()
        # Throw exception, need to clear internal cached host in driver
        self._empty_domain_list = True
        self.driver._vgc_host = None
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.check_for_setup_error)

    @mock.patch('pwd.getpwnam', return_value=1)
    @mock.patch('grp.getgrnam', return_value=1)
    @mock.patch('socket.gethostbyname', return_value='123.123.123.123')
    def test_parameter_storageservers_invalid(self, mock_ghn, mock_grnam,
                                              mock_pwnam):
        """Test exception when the storage servers are invalid/missing."""
        # Should pass
        self.driver.check_for_setup_error()
        # Storage_hosts missing
        self.configuration.hgst_storage_servers = ''
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.check_for_setup_error)
        # missing a : between host and devnode
        self.configuration.hgst_storage_servers = 'stor1,stor2'
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.check_for_setup_error)
        # missing a : between host and devnode
        self.configuration.hgst_storage_servers = 'stor1:gbd0,stor2'
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.check_for_setup_error)
        # Host not in cluster
        self.configuration.hgst_storage_servers = 'stor1:gbd0'
        self._fail_host_storage = True
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.check_for_setup_error)

    def test_update_volume_stats(self):
        """Get cluster space available, should pass."""
        actual = self.driver.get_volume_stats(True)
        self.assertEqual('HGST', actual['vendor_name'])
        self.assertEqual('hgst', actual['storage_protocol'])
        self.assertEqual(90, actual['total_capacity_gb'])
        self.assertEqual(87, actual['free_capacity_gb'])
        self.assertEqual(0, actual['reserved_percentage'])

    def test_update_volume_stats_redundancy(self):
        """Get cluster space available, half-sized - 1 for mirrors."""
        self.configuration.hgst_redundancy = '1'
        actual = self.driver.get_volume_stats(True)
        self.assertEqual('HGST', actual['vendor_name'])
        self.assertEqual('hgst', actual['storage_protocol'])
        self.assertEqual(44, actual['total_capacity_gb'])
        self.assertEqual(43, actual['free_capacity_gb'])
        self.assertEqual(0, actual['reserved_percentage'])

    def test_update_volume_stats_cached(self):
        """Get cached cluster space, should not call executable."""
        self._fail_host_storage = True
        actual = self.driver.get_volume_stats(False)
        self.assertEqual('HGST', actual['vendor_name'])
        self.assertEqual('hgst', actual['storage_protocol'])
        self.assertEqual(90, actual['total_capacity_gb'])
        self.assertEqual(87, actual['free_capacity_gb'])
        self.assertEqual(0, actual['reserved_percentage'])

    def test_update_volume_stats_error(self):
        """Test that when host-storage gives an error, return unknown."""
        self._fail_host_storage = True
        actual = self.driver.get_volume_stats(True)
        self.assertEqual('HGST', actual['vendor_name'])
        self.assertEqual('hgst', actual['storage_protocol'])
        self.assertEqual('unknown', actual['total_capacity_gb'])
        self.assertEqual('unknown', actual['free_capacity_gb'])
        self.assertEqual(0, actual['reserved_percentage'])

    @mock.patch('socket.gethostbyname', return_value='123.123.123.123')
    def test_create_volume(self, mock_ghn):
        """Test volume creation, ensure appropriate size expansion/name."""
        ctxt = context.get_admin_context()
        extra_specs = {}
        type_ref = volume_types.create(ctxt, 'hgst-1', extra_specs)
        volume = {'id': '1', 'name': 'volume1',
                  'display_name': '',
                  'volume_type_id': type_ref['id'],
                  'size': 10}
        ret = self.driver.create_volume(volume)
        expected = {'redundancy': '0', 'group': 'xanadu',
                    'name': 'volume10', 'mode': '0777',
                    'user': 'kane', 'net': 'net1',
                    'storageserver': 'stor1:gbd0,stor2:gbd0,',
                    'size': '12'}
        self.assertDictMatch(expected, self.created)
        # Check the returned provider, note the the provider_id is hashed
        expected_pid = {'provider_id': 'volume10'}
        self.assertDictMatch(expected_pid, ret)

    @mock.patch('socket.gethostbyname', return_value='123.123.123.123')
    def test_create_volume_name_creation_fail(self, mock_ghn):
        """Test volume creation exception when can't make a hashed name."""
        ctxt = context.get_admin_context()
        extra_specs = {}
        type_ref = volume_types.create(ctxt, 'hgst-1', extra_specs)
        volume = {'id': '1', 'name': 'volume1',
                  'display_name': '',
                  'volume_type_id': type_ref['id'],
                  'size': 10}
        self._fail_space_list = True
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.create_volume, volume)

    @mock.patch('socket.gethostbyname', return_value='123.123.123.123')
    def test_create_snapshot(self, mock_ghn):
        """Test creating a snapshot, ensure full data of original copied."""
        # Now snapshot the volume and check commands
        snapshot = {'volume_name': 'volume10',
                    'volume_id': 'xxx', 'display_name': 'snap10',
                    'name': '123abc', 'volume_size': 10, 'id': '123abc',
                    'volume': {'provider_id': 'space10'}}
        ret = self.driver.create_snapshot(snapshot)
        # We must copy entier underlying storage, ~12GB, not just 10GB
        self.assertEqual(11444, self.dd_count)
        # Check space-create command
        expected = {'redundancy': '0', 'group': 'xanadu',
                    'name': snapshot['display_name'], 'mode': '0777',
                    'user': 'kane', 'net': 'net1',
                    'storageserver': 'stor1:gbd0,stor2:gbd0,',
                    'size': '12'}
        self.assertDictMatch(expected, self.created)
        # Check the returned provider
        expected_pid = {'provider_id': 'snap10'}
        self.assertDictMatch(expected_pid, ret)

    @mock.patch('socket.gethostbyname', return_value='123.123.123.123')
    def test_create_cloned_volume(self, mock_ghn):
        """Test creating a clone, ensure full size is copied from original."""
        ctxt = context.get_admin_context()
        extra_specs = {}
        type_ref = volume_types.create(ctxt, 'hgst-1', extra_specs)
        orig = {'id': '1', 'name': 'volume1', 'display_name': '',
                'volume_type_id': type_ref['id'], 'size': 10,
                'provider_id': 'space_orig'}
        clone = {'id': '2', 'name': 'clone1', 'display_name': '',
                 'volume_type_id': type_ref['id'], 'size': 10}
        pid = self.driver.create_cloned_volume(clone, orig)
        # We must copy entier underlying storage, ~12GB, not just 10GB
        self.assertEqual(11444, self.dd_count)
        # Check space-create command
        expected = {'redundancy': '0', 'group': 'xanadu',
                    'name': 'clone1', 'mode': '0777',
                    'user': 'kane', 'net': 'net1',
                    'storageserver': 'stor1:gbd0,stor2:gbd0,',
                    'size': '12'}
        self.assertDictMatch(expected, self.created)
        # Check the returned provider
        expected_pid = {'provider_id': 'clone1'}
        self.assertDictMatch(expected_pid, pid)

    @mock.patch('socket.gethostbyname', return_value='123.123.123.123')
    def test_add_cinder_apphosts_fails(self, mock_ghn):
        """Test exception when set-apphost can't connect volume to host."""
        ctxt = context.get_admin_context()
        extra_specs = {}
        type_ref = volume_types.create(ctxt, 'hgst-1', extra_specs)
        orig = {'id': '1', 'name': 'volume1', 'display_name': '',
                'volume_type_id': type_ref['id'], 'size': 10,
                'provider_id': 'space_orig'}
        clone = {'id': '2', 'name': 'clone1', 'display_name': '',
                 'volume_type_id': type_ref['id'], 'size': 10}
        self._fail_set_apphosts = True
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.create_cloned_volume, clone, orig)

    @mock.patch('socket.gethostbyname', return_value='123.123.123.123')
    def test_create_volume_from_snapshot(self, mock_ghn):
        """Test creating volume from snapshot, ensure full space copy."""
        ctxt = context.get_admin_context()
        extra_specs = {}
        type_ref = volume_types.create(ctxt, 'hgst-1', extra_specs)
        snap = {'id': '1', 'name': 'volume1', 'display_name': '',
                'volume_type_id': type_ref['id'], 'size': 10,
                'provider_id': 'space_orig'}
        volume = {'id': '2', 'name': 'volume2', 'display_name': '',
                  'volume_type_id': type_ref['id'], 'size': 10}
        pid = self.driver.create_volume_from_snapshot(volume, snap)
        # We must copy entier underlying storage, ~12GB, not just 10GB
        self.assertEqual(11444, self.dd_count)
        # Check space-create command
        expected = {'redundancy': '0', 'group': 'xanadu',
                    'name': 'volume2', 'mode': '0777',
                    'user': 'kane', 'net': 'net1',
                    'storageserver': 'stor1:gbd0,stor2:gbd0,',
                    'size': '12'}
        self.assertDictMatch(expected, self.created)
        # Check the returned provider
        expected_pid = {'provider_id': 'volume2'}
        self.assertDictMatch(expected_pid, pid)

    @mock.patch('socket.gethostbyname', return_value='123.123.123.123')
    def test_create_volume_blocked(self, mock_ghn):
        """Test volume creation where only initial space-create is blocked.

        This should actually pass because we are blocked byt return an error
        in request-cancel, meaning that it got unblocked before we could kill
        the space request.
        """
        ctxt = context.get_admin_context()
        extra_specs = {}
        type_ref = volume_types.create(ctxt, 'hgst-1', extra_specs)
        volume = {'id': '1', 'name': 'volume1',
                  'display_name': '',
                  'volume_type_id': type_ref['id'],
                  'size': 10}
        self._return_blocked = 1  # Block & fail cancel => create succeeded
        ret = self.driver.create_volume(volume)
        expected = {'redundancy': '0', 'group': 'xanadu',
                    'name': 'volume10', 'mode': '0777',
                    'user': 'kane', 'net': 'net1',
                    'storageserver': 'stor1:gbd0,stor2:gbd0,',
                    'size': '12'}
        self.assertDictMatch(expected, self.created)
        # Check the returned provider
        expected_pid = {'provider_id': 'volume10'}
        self.assertDictMatch(expected_pid, ret)
        self.assertTrue(self._request_cancel)

    @mock.patch('socket.gethostbyname', return_value='123.123.123.123')
    def test_create_volume_blocked_and_fail(self, mock_ghn):
        """Test volume creation where space-create blocked permanently.

        This should fail because the initial create was blocked and the
        request-cancel succeeded, meaning the create operation never
        completed.
        """
        ctxt = context.get_admin_context()
        extra_specs = {}
        type_ref = volume_types.create(ctxt, 'hgst-1', extra_specs)
        volume = {'id': '1', 'name': 'volume1',
                  'display_name': '',
                  'volume_type_id': type_ref['id'],
                  'size': 10}
        self._return_blocked = 2  # Block & pass cancel => create failed. :(
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.create_volume, volume)
        self.assertTrue(self._request_cancel)

    def test_delete_volume(self):
        """Test deleting existing volume, ensure proper name used."""
        ctxt = context.get_admin_context()
        extra_specs = {}
        type_ref = volume_types.create(ctxt, 'hgst-1', extra_specs)
        volume = {'id': '1', 'name': 'volume1',
                  'display_name': '',
                  'volume_type_id': type_ref['id'],
                  'size': 10,
                  'provider_id': 'volume10'}
        self.driver.delete_volume(volume)
        expected = {'name': 'volume10'}
        self.assertDictMatch(expected, self.deleted)

    def test_delete_volume_failure_modes(self):
        """Test cases where space-delete fails, but OS delete is still OK."""
        ctxt = context.get_admin_context()
        extra_specs = {}
        type_ref = volume_types.create(ctxt, 'hgst-1', extra_specs)
        volume = {'id': '1', 'name': 'volume1',
                  'display_name': '',
                  'volume_type_id': type_ref['id'],
                  'size': 10,
                  'provider_id': 'volume10'}
        self._fail_space_delete = True
        # This should not throw an exception, space-delete failure not problem
        self.driver.delete_volume(volume)
        self._fail_space_delete = False
        volume['provider_id'] = None
        # This should also not throw an exception
        self.driver.delete_volume(volume)

    def test_delete_snapshot(self):
        """Test deleting a snapshot, ensure proper name is removed."""
        ctxt = context.get_admin_context()
        extra_specs = {}
        type_ref = volume_types.create(ctxt, 'hgst-1', extra_specs)
        snapshot = {'id': '1', 'name': 'volume1',
                    'display_name': '',
                    'volume_type_id': type_ref['id'],
                    'size': 10,
                    'provider_id': 'snap10'}
        self.driver.delete_snapshot(snapshot)
        expected = {'name': 'snap10'}
        self.assertDictMatch(expected, self.deleted)

    def test_extend_volume(self):
        """Test extending a volume, check the size in GB vs. GiB."""
        ctxt = context.get_admin_context()
        extra_specs = {}
        type_ref = volume_types.create(ctxt, 'hgst-1', extra_specs)
        volume = {'id': '1', 'name': 'volume1',
                  'display_name': '',
                  'volume_type_id': type_ref['id'],
                  'size': 10,
                  'provider_id': 'volume10'}
        self.extended = {'name': '', 'size': '0',
                         'storageserver': ''}
        self.driver.extend_volume(volume, 12)
        expected = {'name': 'volume10', 'size': '2',
                    'storageserver': 'stor1:gbd0,stor2:gbd0,'}
        self.assertDictMatch(expected, self.extended)

    def test_extend_volume_noextend(self):
        """Test extending a volume where Space does not need to be enlarged.

        Because Spaces are generated somewhat larger than the requested size
        from OpenStack due to the base10(HGST)/base2(OS) mismatch, they can
        sometimes be larger than requested from OS.  In that case a
        volume_extend may actually be a noop since the volume is already large
        enough to satisfy OS's request.
        """
        ctxt = context.get_admin_context()
        extra_specs = {}
        type_ref = volume_types.create(ctxt, 'hgst-1', extra_specs)
        volume = {'id': '1', 'name': 'volume1',
                  'display_name': '',
                  'volume_type_id': type_ref['id'],
                  'size': 10,
                  'provider_id': 'volume10'}
        self.extended = {'name': '', 'size': '0',
                         'storageserver': ''}
        self.driver.extend_volume(volume, 10)
        expected = {'name': '', 'size': '0',
                    'storageserver': ''}
        self.assertDictMatch(expected, self.extended)

    def test_space_list_fails(self):
        """Test exception is thrown when we can't call space-list."""
        ctxt = context.get_admin_context()
        extra_specs = {}
        type_ref = volume_types.create(ctxt, 'hgst-1', extra_specs)
        volume = {'id': '1', 'name': 'volume1',
                  'display_name': '',
                  'volume_type_id': type_ref['id'],
                  'size': 10,
                  'provider_id': 'volume10'}
        self.extended = {'name': '', 'size': '0',
                         'storageserver': ''}
        self._fail_space_list = True
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.extend_volume, volume, 12)

    def test_cli_error_not_blocked(self):
        """Test the _blocked handler's handlinf of a non-blocked error.

        The _handle_blocked handler is called on any process errors in the
        code.  If the error was not caused by a blocked command condition
        (syntax error, out of space, etc.) then it should just throw the
        exception and not try and retry the command.
        """
        ctxt = context.get_admin_context()
        extra_specs = {}
        type_ref = volume_types.create(ctxt, 'hgst-1', extra_specs)
        volume = {'id': '1', 'name': 'volume1',
                  'display_name': '',
                  'volume_type_id': type_ref['id'],
                  'size': 10,
                  'provider_id': 'volume10'}
        self.extended = {'name': '', 'size': '0',
                         'storageserver': ''}
        self._fail_extend = True
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.extend_volume, volume, 12)
        self.assertFalse(self._request_cancel)

    @mock.patch('socket.gethostbyname', return_value='123.123.123.123')
    def test_initialize_connection(self, moch_ghn):
        """Test that the connection_info for Nova makes sense."""
        volume = {'name': '123', 'provider_id': 'spacey'}
        conn = self.driver.initialize_connection(volume, None)
        expected = {'name': 'spacey', 'noremovehost': 'thisserver'}
        self.assertDictMatch(expected, conn['data'])

# Below are some command outputs we emulate
IP_OUTPUT = """
3: em2: <BROADCAST,MULTICAST,UP,LOWER_UP> mtu 1500 qdisc mq state
    link/ether 00:25:90:d9:18:09 brd ff:ff:ff:ff:ff:ff
    inet 192.168.0.23/24 brd 192.168.0.255 scope global em2
       valid_lft forever preferred_lft forever
    inet6 fe80::225:90ff:fed9:1809/64 scope link
       valid_lft forever preferred_lft forever
1: lo: <LOOPBACK,UP,LOWER_UP> mtu 65536 qdisc noqueue state UNKNOWN
    link/loopback 00:00:00:00:00:00 brd 00:00:00:00:00:00
    inet 123.123.123.123/8 scope host lo
       valid_lft forever preferred_lft forever
    inet 169.254.169.254/32 scope link lo
       valid_lft forever preferred_lft forever
    inet6 ::1/128 scope host
       valid_lft forever preferred_lft forever
2: em1: <BROADCAST,MULTICAST,UP,LOWER_UP> mtu 1500 qdisc mq master
    link/ether 00:25:90:d9:18:08 brd ff:ff:ff:ff:ff:ff
    inet6 fe80::225:90ff:fed9:1808/64 scope link
       valid_lft forever preferred_lft forever
"""

HGST_HOST_STORAGE = """
{
  "hostStatus": [
    {
      "node": "tm33.virident.info",
      "up": true,
      "isManager": true,
      "cardStatus": [
        {
          "cardName": "/dev/sda3",
          "cardSerialNumber": "002f09b4037a9d521c007ee4esda3",
          "cardStatus": "Good",
          "cardStateDetails": "Normal",
          "cardActionRequired": "",
          "cardTemperatureC": 0,
          "deviceType": "Generic",
          "cardTemperatureState": "Safe",
          "partitionStatus": [
            {
              "partName": "/dev/gbd0",
              "partitionState": "READY",
              "usableCapacityBytes": 98213822464,
              "totalReadBytes": 0,
              "totalWriteBytes": 0,
              "remainingLifePCT": 100,
              "flashReservesLeftPCT": 100,
              "fmc": true,
              "vspaceCapacityAvailable": 94947041280,
              "vspaceReducedCapacityAvailable": 87194279936,
              "_partitionID": "002f09b4037a9d521c007ee4esda3:0",
              "_usedSpaceBytes": 3266781184,
              "_enabledSpaceBytes": 3266781184,
              "_disabledSpaceBytes": 0
            }
          ]
        }
      ],
      "driverStatus": {
        "vgcdriveDriverLoaded": true,
        "vhaDriverLoaded": true,
        "vcacheDriverLoaded": true,
        "vlvmDriverLoaded": true,
        "ipDataProviderLoaded": true,
        "ibDataProviderLoaded": false,
        "driverUptimeSecs": 4800,
        "rVersion": "20368.d55ec22.master"
      },
      "totalCapacityBytes": 98213822464,
      "totalUsedBytes": 3266781184,
      "totalEnabledBytes": 3266781184,
      "totalDisabledBytes": 0
    },
    {
      "node": "tm32.virident.info",
      "up": true,
      "isManager": false,
      "cardStatus": [],
      "driverStatus": {
        "vgcdriveDriverLoaded": true,
        "vhaDriverLoaded": true,
        "vcacheDriverLoaded": true,
        "vlvmDriverLoaded": true,
        "ipDataProviderLoaded": true,
        "ibDataProviderLoaded": false,
        "driverUptimeSecs": 0,
        "rVersion": "20368.d55ec22.master"
      },
      "totalCapacityBytes": 0,
      "totalUsedBytes": 0,
      "totalEnabledBytes": 0,
      "totalDisabledBytes": 0
    }
  ],
  "totalCapacityBytes": 98213822464,
  "totalUsedBytes": 3266781184,
  "totalEnabledBytes": 3266781184,
  "totalDisabledBytes": 0
}
"""

HGST_SPACE_JSON = """
{
  "resources": [
    {
      "resourceType": "vLVM-L",
      "resourceID": "vLVM-L:698cdb43-54da-863e-1699-294a080ce4db",
      "state": "OFFLINE",
      "instanceStates": {},
      "redundancy": 0,
      "sizeBytes": 12000000000,
      "name": "volume10",
      "nodes": [],
      "networks": [
        "net1"
      ],
      "components": [
        {
          "resourceType": "vLVM-S",
          "resourceID": "vLVM-S:698cdb43-54da-863e-eb10-6275f47b8ed2",
          "redundancy": 0,
          "order": 0,
          "sizeBytes": 12000000000,
          "numStripes": 1,
          "stripeSizeBytes": null,
          "name": "volume10s00",
          "state": "OFFLINE",
          "instanceStates": {},
          "components": [
            {
              "name": "volume10h00",
              "resourceType": "vHA",
              "resourceID": "vHA:3e86da54-40db-8c69-0300-0000ac10476e",
              "redundancy": 0,
              "sizeBytes": 12000000000,
              "state": "GOOD",
              "components": [
                {
                  "name": "volume10h00",
                  "vspaceType": "vHA",
                  "vspaceRole": "primary",
                  "storageObjectID": "vHA:3e86da54-40db-8c69--18130019e486",
                  "state": "Disconnected (DCS)",
                  "node": "tm33.virident.info",
                  "partName": "/dev/gbd0"
                }
              ],
              "crState": "GOOD"
            },
            {
              "name": "volume10v00",
              "resourceType": "vShare",
              "resourceID": "vShare:3f86da54-41db-8c69-0300-ecf4bbcc14cc",
              "redundancy": 0,
              "order": 0,
              "sizeBytes": 12000000000,
              "state": "GOOD",
              "components": [
                {
                  "name": "volume10v00",
                  "vspaceType": "vShare",
                  "vspaceRole": "target",
                  "storageObjectID": "vShare:3f86da54-41db-8c64bbcc14cc:T",
                  "state": "Started",
                  "node": "tm33.virident.info",
                  "partName": "/dev/gbd0_volume10h00"
                }
              ]
            }
          ]
        }
      ],
      "_size": "12GB",
      "_state": "OFFLINE",
      "_ugm": "",
      "_nets": "net1",
      "_hosts": "tm33.virident.info(12GB,NC)",
      "_ahosts": "",
      "_shosts": "tm33.virident.info(12GB)",
      "_name": "volume10",
      "_node": "",
      "_type": "vLVM-L",
      "_detail": "vLVM-L:698cdb43-54da-863e-1699-294a080ce4db",
      "_device": ""
    }
  ]
}
"""

NETWORK_LIST = """
Network Name Type    Flags          Description
------------ ---- ---------- ------------------------
net1         IPv4 autoConfig 192.168.0.0/24 1Gb/s
net2         IPv4 autoConfig 192.168.10.0/24 10Gb/s
"""

DD_OUTPUT = """
1+0 records in
1+0 records out
1024 bytes (1.0 kB) copied, 0.000427529 s, 2.4 MB/s
"""



# Copyright (C) 2014, Hitachi, Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
iSCSI Cinder volume driver for Hitachi storage.

"""

import os
import threading

from oslo_config import cfg
from oslo_log import log as logging
import six

from cinder import exception
from cinder.i18n import _LE, _LI
from cinder import utils
import cinder.volume.driver
from cinder.volume.drivers.hitachi import hbsd_basiclib as basic_lib
from cinder.volume.drivers.hitachi import hbsd_common as common

LOG = logging.getLogger(__name__)

CHAP_METHOD = ('None', 'CHAP None', 'CHAP')

volume_opts = [
    cfg.BoolOpt('hitachi_add_chap_user',
                default=False,
                help='Add CHAP user'),
    cfg.StrOpt('hitachi_auth_method',
               help='iSCSI authentication method'),
    cfg.StrOpt('hitachi_auth_user',
               default='%sCHAP-user' % basic_lib.NAME_PREFIX,
               help='iSCSI authentication username'),
    cfg.StrOpt('hitachi_auth_password',
               default='%sCHAP-password' % basic_lib.NAME_PREFIX,
               help='iSCSI authentication password', secret=True),
]

CONF = cfg.CONF
CONF.register_opts(volume_opts)


class HBSDISCSIDriver(cinder.volume.driver.ISCSIDriver):
    VERSION = common.VERSION

    def __init__(self, *args, **kwargs):
        os.environ['LANG'] = 'C'
        super(HBSDISCSIDriver, self).__init__(*args, **kwargs)
        self.db = kwargs.get('db')
        self.common = None
        self.configuration.append_config_values(common.volume_opts)
        self._stats = {}
        self.context = None
        self.do_setup_status = threading.Event()

    def _check_param(self):
        self.configuration.append_config_values(volume_opts)
        if (self.configuration.hitachi_auth_method and
                self.configuration.hitachi_auth_method not in CHAP_METHOD):
            raise exception.HBSDError(
                message=basic_lib.output_err(601, param='hitachi_auth_method'))
        if self.configuration.hitachi_auth_method == 'None':
            self.configuration.hitachi_auth_method = None
        for opt in volume_opts:
            getattr(self.configuration, opt.name)

    def check_param(self):
        try:
            self.common.check_param()
            self._check_param()
        except exception.HBSDError:
            raise
        except Exception as ex:
            raise exception.HBSDError(
                message=basic_lib.output_err(601, param=six.text_type(ex)))

    def output_param_to_log(self):
        lock = basic_lib.get_process_lock(self.common.system_lock_file)

        with lock:
            self.common.output_param_to_log('iSCSI')
            for opt in volume_opts:
                if not opt.secret:
                    value = getattr(self.configuration, opt.name)
                    LOG.info(_LI('\t%(name)-35s : %(value)s'),
                             {'name': opt.name, 'value': value})

    def _delete_lun_iscsi(self, hostgroups, ldev):
        try:
            self.common.command.comm_delete_lun_iscsi(hostgroups, ldev)
        except exception.HBSDNotFound:
            LOG.warning(basic_lib.set_msg(301, ldev=ldev))

    def _add_target(self, hostgroups, ldev):
        self.common.add_lun('autargetmap', hostgroups, ldev)

    def _add_initiator(self, hgs, port, gid, host_iqn):
        self.common.command.comm_add_initiator(port, gid, host_iqn)
        hgs.append({'port': port, 'gid': int(gid), 'detected': True})
        LOG.debug("Create iSCSI target for %s", hgs)

    def _get_unused_gid_iscsi(self, port):
        group_range = self.configuration.hitachi_group_range
        if not group_range:
            group_range = basic_lib.DEFAULT_GROUP_RANGE
        return self.common.command.get_unused_gid_iscsi(group_range, port)

    def _delete_iscsi_target(self, port, target_no, target_alias):
        ret, _stdout, _stderr = self.common.command.delete_iscsi_target(
            port, target_no, target_alias)
        if ret:
            LOG.warning(basic_lib.set_msg(
                307, port=port, tno=target_no, alias=target_alias))

    def _delete_chap_user(self, port):
        ret, _stdout, _stderr = self.common.command.delete_chap_user(port)
        if ret:
            LOG.warning(basic_lib.set_msg(
                303, user=self.configuration.hitachi_auth_user))

    def _get_hostgroup_info_iscsi(self, hgs, host_iqn):
        return self.common.command.comm_get_hostgroup_info_iscsi(
            hgs, host_iqn, self.configuration.hitachi_target_ports)

    def _discovery_iscsi_target(self, hostgroups):
        for hostgroup in hostgroups:
            ip_addr, ip_port = self.common.command.comm_get_iscsi_ip(
                hostgroup['port'])
            target_iqn = self.common.command.comm_get_target_iqn(
                hostgroup['port'], hostgroup['gid'])
            hostgroup['ip_addr'] = ip_addr
            hostgroup['ip_port'] = ip_port
            hostgroup['target_iqn'] = target_iqn
            LOG.debug("ip_addr=%(addr)s ip_port=%(port)s target_iqn=%(iqn)s",
                      {'addr': ip_addr, 'port': ip_port, 'iqn': target_iqn})

    def _fill_groups(self, hgs, ports, target_iqn, target_alias, add_iqn):
        for port in ports:
            added_hostgroup = False
            added_user = False
            LOG.debug('Create target (hgs: %(hgs)s port: %(port)s '
                      'target_iqn: %(tiqn)s target_alias: %(alias)s '
                      'add_iqn: %(aiqn)s)',
                      {'hgs': hgs, 'port': port, 'tiqn': target_iqn,
                       'alias': target_alias, 'aiqn': add_iqn})
            gid = self.common.command.get_gid_from_targetiqn(
                target_iqn, target_alias, port)
            if gid is None:
                for retry_cnt in basic_lib.DEFAULT_TRY_RANGE:
                    gid = None
                    try:
                        gid = self._get_unused_gid_iscsi(port)
                        self.common.command.comm_add_hostgrp_iscsi(
                            port, gid, target_alias, target_iqn)
                        added_hostgroup = True
                    except exception.HBSDNotFound:
                        LOG.warning(basic_lib.set_msg(312, resource='GID'))
                        continue
                    except Exception as ex:
                        LOG.warning(basic_lib.set_msg(
                            309, port=port, alias=target_alias,
                            reason=ex))
                        break
                    else:
                        LOG.debug('Completed to add target'
                                  '(port: %(port)s gid: %(gid)d)',
                                  {'port': port, 'gid': gid})
                        break
            if gid is None:
                LOG.error(_LE('Failed to add target(port: %s)'), port)
                continue
            try:
                if added_hostgroup:
                    if self.configuration.hitachi_auth_method:
                        added_user = self.common.command.set_chap_authention(
                            port, gid)
                    self.common.command.comm_set_hostgrp_reportportal(
                        port, target_alias)
                self._add_initiator(hgs, port, gid, add_iqn)
            except Exception as ex:
                LOG.warning(basic_lib.set_msg(
                    316, port=port, reason=ex))
                if added_hostgroup:
                    if added_user:
                        self._delete_chap_user(port)
                    self._delete_iscsi_target(port, gid, target_alias)

    def add_hostgroup_core(self, hgs, ports, target_iqn,
                           target_alias, add_iqn):
        if ports:
            self._fill_groups(hgs, ports, target_iqn, target_alias, add_iqn)

    def add_hostgroup_master(self, hgs, master_iqn, host_ip, security_ports):
        target_ports = self.configuration.hitachi_target_ports
        group_request = self.configuration.hitachi_group_request
        target_alias = '%s%s' % (basic_lib.NAME_PREFIX, host_ip)
        if target_ports and group_request:
            target_iqn = '%s.target' % master_iqn

            diff_ports = []
            for port in security_ports:
                for hostgroup in hgs:
                    if hostgroup['port'] == port:
                        break
                else:
                    diff_ports.append(port)

            self.add_hostgroup_core(hgs, diff_ports, target_iqn,
                                    target_alias, master_iqn)
        if not hgs:
            raise exception.HBSDError(message=basic_lib.output_err(649))

    def add_hostgroup(self):
        properties = utils.brick_get_connector_properties()
        if 'initiator' not in properties:
            raise exception.HBSDError(
                message=basic_lib.output_err(650, resource='HBA'))
        LOG.debug("initiator: %s", properties['initiator'])
        hostgroups = []
        security_ports = self._get_hostgroup_info_iscsi(
            hostgroups, properties['initiator'])
        self.add_hostgroup_master(hostgroups, properties['initiator'],
                                  properties['ip'], security_ports)

    def _get_properties(self, volume, hostgroups):
        conf = self.configuration
        properties = {}
        self._discovery_iscsi_target(hostgroups)
        hostgroup = hostgroups[0]

        properties['target_discovered'] = True
        properties['target_portal'] = "%s:%s" % (hostgroup['ip_addr'],
                                                 hostgroup['ip_port'])
        properties['target_iqn'] = hostgroup['target_iqn']
        properties['target_lun'] = hostgroup['lun']

        if conf.hitachi_auth_method:
            properties['auth_method'] = 'CHAP'
            properties['auth_username'] = conf.hitachi_auth_user
            properties['auth_password'] = conf.hitachi_auth_password

        return properties

    def do_setup(self, context):
        self.context = context
        self.common = common.HBSDCommon(self.configuration, self,
                                        context, self.db)

        self.check_param()

        self.common.create_lock_file()

        self.common.command.connect_storage()

        lock = basic_lib.get_process_lock(self.common.service_lock_file)
        with lock:
            self.add_hostgroup()

        self.output_param_to_log()
        self.do_setup_status.set()

    def check_for_setup_error(self):
        pass

    def extend_volume(self, volume, new_size):
        self.do_setup_status.wait()
        self.common.extend_volume(volume, new_size)

    def get_volume_stats(self, refresh=False):
        if refresh:
            if self.do_setup_status.isSet():
                self.common.output_backend_available_once()
                _stats = self.common.update_volume_stats("iSCSI")
                if _stats:
                    self._stats = _stats
        return self._stats

    def create_volume(self, volume):
        self.do_setup_status.wait()
        metadata = self.common.create_volume(volume)
        return metadata

    def delete_volume(self, volume):
        self.do_setup_status.wait()
        self.common.delete_volume(volume)

    def create_snapshot(self, snapshot):
        self.do_setup_status.wait()
        metadata = self.common.create_snapshot(snapshot)
        return metadata

    def delete_snapshot(self, snapshot):
        self.do_setup_status.wait()
        self.common.delete_snapshot(snapshot)

    def create_cloned_volume(self, volume, src_vref):
        self.do_setup_status.wait()
        metadata = self.common.create_cloned_volume(volume, src_vref)
        return metadata

    def create_volume_from_snapshot(self, volume, snapshot):
        self.do_setup_status.wait()
        metadata = self.common.create_volume_from_snapshot(volume, snapshot)
        return metadata

    def _initialize_connection(self, ldev, connector, src_hgs=None):
        LOG.debug("Call _initialize_connection "
                  "(config_group: %(group)s ldev: %(ldev)d)",
                  {'group': self.configuration.config_group, 'ldev': ldev})
        if src_hgs:
            hostgroups = src_hgs[:]
        else:
            hostgroups = []
            security_ports = self._get_hostgroup_info_iscsi(
                hostgroups, connector['initiator'])
            self.add_hostgroup_master(hostgroups, connector['initiator'],
                                      connector['ip'], security_ports)

        self._add_target(hostgroups, ldev)

        return hostgroups

    def initialize_connection(self, volume, connector):
        self.do_setup_status.wait()
        ldev = self.common.get_ldev(volume)
        if ldev is None:
            raise exception.HBSDError(
                message=basic_lib.output_err(619, volume_id=volume['id']))
        self.common.add_volinfo(ldev, volume['id'])
        with self.common.volume_info[ldev]['lock'],\
                self.common.volume_info[ldev]['in_use']:
            hostgroups = self._initialize_connection(ldev, connector)
            protocol = 'iscsi'
            properties = self._get_properties(volume, hostgroups)
            LOG.debug('Initialize volume_info: %s',
                      self.common.volume_info)

        LOG.debug('HFCDrv: properties=%s', properties)
        return {
            'driver_volume_type': protocol,
            'data': properties
        }

    def _terminate_connection(self, ldev, connector, src_hgs):
        LOG.debug("Call _terminate_connection(config_group: %s)",
                  self.configuration.config_group)
        hostgroups = src_hgs[:]
        self._delete_lun_iscsi(hostgroups, ldev)

        LOG.debug("*** _terminate_ ***")

    def terminate_connection(self, volume, connector, **kwargs):
        self.do_setup_status.wait()
        ldev = self.common.get_ldev(volume)
        if ldev is None:
            LOG.warning(basic_lib.set_msg(302, volume_id=volume['id']))
            return

        if 'initiator' not in connector:
            raise exception.HBSDError(
                message=basic_lib.output_err(650, resource='HBA'))

        hostgroups = []
        self._get_hostgroup_info_iscsi(hostgroups,
                                       connector['initiator'])
        if not hostgroups:
            raise exception.HBSDError(message=basic_lib.output_err(649))

        self.common.add_volinfo(ldev, volume['id'])
        with self.common.volume_info[ldev]['lock'],\
                self.common.volume_info[ldev]['in_use']:
            self._terminate_connection(ldev, connector, hostgroups)

    def create_export(self, context, volume, connector):
        pass

    def ensure_export(self, context, volume):
        pass

    def remove_export(self, context, volume):
        pass

    def pair_initialize_connection(self, unused_ldev):
        pass

    def pair_terminate_connection(self, unused_ldev):
        pass

    def copy_volume_to_image(self, context, volume, image_service, image_meta):
        self.do_setup_status.wait()
        if volume['volume_attachment']:
            desc = 'volume %s' % volume['id']
            raise exception.HBSDError(
                message=basic_lib.output_err(660, desc=desc))
        super(HBSDISCSIDriver, self).copy_volume_to_image(context, volume,
                                                          image_service,
                                                          image_meta)

    def manage_existing(self, volume, existing_ref):
        return self.common.manage_existing(volume, existing_ref)

    def manage_existing_get_size(self, volume, existing_ref):
        self.do_setup_status.wait()
        return self.common.manage_existing_get_size(volume, existing_ref)

    def unmanage(self, volume):
        self.do_setup_status.wait()
        self.common.unmanage(volume)



# Copyright 2013 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Handles all requests to Nova.
"""


from novaclient import client as nova_client
from novaclient import exceptions as nova_exceptions
from novaclient import service_catalog
from oslo_config import cfg
from oslo_log import log as logging
from requests import exceptions as request_exceptions

from cinder import context as ctx
from cinder.db import base
from cinder import exception

nova_opts = [
    cfg.StrOpt('nova_catalog_info',
               default='compute:Compute Service:publicURL',
               help='Match this value when searching for nova in the '
                    'service catalog. Format is: separated values of '
                    'the form: '
                    '<service_type>:<service_name>:<endpoint_type>'),
    cfg.StrOpt('nova_catalog_admin_info',
               default='compute:Compute Service:adminURL',
               help='Same as nova_catalog_info, but for admin endpoint.'),
    cfg.StrOpt('nova_endpoint_template',
               help='Override service catalog lookup with template for nova '
                    'endpoint e.g. http://localhost:8774/v2/%(project_id)s'),
    cfg.StrOpt('nova_endpoint_admin_template',
               help='Same as nova_endpoint_template, but for admin endpoint.'),
    cfg.StrOpt('os_region_name',
               help='Region name of this node'),
    cfg.StrOpt('nova_ca_certificates_file',
               help='Location of ca certificates file to use for nova client '
                    'requests.'),
    cfg.BoolOpt('nova_api_insecure',
                default=False,
                help='Allow to perform insecure SSL requests to nova'),
]

CONF = cfg.CONF
CONF.register_opts(nova_opts)

LOG = logging.getLogger(__name__)

# TODO(e0ne): Make Nova version configurable in Mitaka.
NOVA_API_VERSION = 2

nova_extensions = [ext for ext in nova_client.discover_extensions(2)
                   if ext.name in ("assisted_volume_snapshots",
                                   "list_extensions")]


def novaclient(context, admin_endpoint=False, privileged_user=False,
               timeout=None):
    """Returns a Nova client

    @param admin_endpoint: If True, use the admin endpoint template from
        configuration ('nova_endpoint_admin_template' and 'nova_catalog_info')
    @param privileged_user: If True, use the account from configuration
        (requires 'os_privileged_user_name', 'os_privileged_user_password' and
        'os_privileged_user_tenant' to be set)
    @param timeout: Number of seconds to wait for an answer before raising a
        Timeout exception (None to disable)
    """
    # FIXME: the novaclient ServiceCatalog object is mis-named.
    #        It actually contains the entire access blob.
    # Only needed parts of the service catalog are passed in, see
    # nova/context.py.
    compat_catalog = {
        'access': {'serviceCatalog': context.service_catalog or []}
    }
    sc = service_catalog.ServiceCatalog(compat_catalog)

    nova_endpoint_template = CONF.nova_endpoint_template
    nova_catalog_info = CONF.nova_catalog_info

    if admin_endpoint:
        nova_endpoint_template = CONF.nova_endpoint_admin_template
        nova_catalog_info = CONF.nova_catalog_admin_info
    service_type, service_name, endpoint_type = nova_catalog_info.split(':')

    # Extract the region if set in configuration
    if CONF.os_region_name:
        region_filter = {'attr': 'region', 'filter_value': CONF.os_region_name}
    else:
        region_filter = {}

    if privileged_user and CONF.os_privileged_user_name:
        context = ctx.RequestContext(
            CONF.os_privileged_user_name, None,
            auth_token=CONF.os_privileged_user_password,
            project_name=CONF.os_privileged_user_tenant,
            service_catalog=context.service_catalog)

        # When privileged_user is used, it needs to authenticate to Keystone
        # before querying Nova, so we set auth_url to the identity service
        # endpoint.
        if CONF.os_privileged_user_auth_url:
            url = CONF.os_privileged_user_auth_url
        else:
            # We then pass region_name, endpoint_type, etc. to the
            # Client() constructor so that the final endpoint is
            # chosen correctly.
            url = sc.url_for(service_type='identity',
                             endpoint_type=endpoint_type,
                             **region_filter)

        LOG.debug('Creating a Nova client using "%s" user',
                  CONF.os_privileged_user_name)
    else:
        if nova_endpoint_template:
            url = nova_endpoint_template % context.to_dict()
        else:
            url = sc.url_for(service_type=service_type,
                             service_name=service_name,
                             endpoint_type=endpoint_type,
                             **region_filter)

        LOG.debug('Nova client connection created using URL: %s', url)

    c = nova_client.Client(NOVA_API_VERSION,
                           context.user_id,
                           context.auth_token,
                           context.project_name,
                           auth_url=url,
                           insecure=CONF.nova_api_insecure,
                           timeout=timeout,
                           region_name=CONF.os_region_name,
                           endpoint_type=endpoint_type,
                           cacert=CONF.nova_ca_certificates_file,
                           extensions=nova_extensions)

    if not privileged_user:
        # noauth extracts user_id:project_id from auth_token
        c.client.auth_token = (context.auth_token or '%s:%s'
                               % (context.user_id, context.project_id))
        c.client.management_url = url
    return c


class API(base.Base):
    """API for interacting with novaclient."""

    def has_extension(self, context, extension, timeout=None):
        try:
            nova_exts = novaclient(context).list_extensions.show_all()
        except request_exceptions.Timeout:
            raise exception.APITimeout(service='Nova')
        return extension in [e.name for e in nova_exts]

    def update_server_volume(self, context, server_id, attachment_id,
                             new_volume_id):
        novaclient(context).volumes.update_server_volume(server_id,
                                                         attachment_id,
                                                         new_volume_id)

    def create_volume_snapshot(self, context, volume_id, create_info):
        nova = novaclient(context, admin_endpoint=True, privileged_user=True)

        # pylint: disable-msg=E1101
        nova.assisted_volume_snapshots.create(
            volume_id,
            create_info=create_info)

    def delete_volume_snapshot(self, context, snapshot_id, delete_info):
        nova = novaclient(context, admin_endpoint=True, privileged_user=True)

        # pylint: disable-msg=E1101
        nova.assisted_volume_snapshots.delete(
            snapshot_id,
            delete_info=delete_info)

    def get_server(self, context, server_id, privileged_user=False,
                   timeout=None):
        try:
            return novaclient(context, privileged_user=privileged_user,
                              timeout=timeout).servers.get(server_id)
        except nova_exceptions.NotFound:
            raise exception.ServerNotFound(uuid=server_id)
        except request_exceptions.Timeout:
            raise exception.APITimeout(service='Nova')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
""" Tests for manage_existing TaskFlow """

from cinder import context
from cinder import test
from cinder.tests.unit import fake_volume
from cinder.tests.unit.volume.flows import fake_volume_api
from cinder.volume.flows.api import manage_existing


class ManageVolumeFlowTestCase(test.TestCase):

    def setUp(self):
        super(ManageVolumeFlowTestCase, self).setUp()
        self.ctxt = context.get_admin_context()
        self.counter = float(0)

    def test_cast_manage_existing(self):
        volume = fake_volume.fake_volume_type_obj(self.ctxt)

        spec = {
            'name': 'name',
            'description': 'description',
            'host': 'host',
            'ref': 'ref',
            'volume_type': 'volume_type',
            'metadata': 'metadata',
            'availability_zone': 'availability_zone',
            'bootable': 'bootable',
            'volume_id': volume.id,
        }

        # Fake objects assert specs
        task = manage_existing.ManageCastTask(
            fake_volume_api.FakeSchedulerRpcAPI(spec, self),
            fake_volume_api.FakeDb())

        create_what = spec.copy()
        create_what.update({'volume': volume})
        create_what.pop('volume_id')
        task.execute(self.ctxt, **create_what)



#    Copyright 2013 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

from oslo_config import cfg
from sqlalchemy import MetaData, Table

# Get default values via config.  The defaults will either
# come from the default values set in the quota option
# configuration or via cinder.conf if the user has configured
# default values for quotas there.
CONF = cfg.CONF
CONF.import_opt('quota_volumes', 'cinder.quota')
CONF.import_opt('quota_snapshots', 'cinder.quota')
CONF.import_opt('quota_gigabytes', 'cinder.quota')

CLASS_NAME = 'default'
CREATED_AT = datetime.datetime.now()  # noqa


def upgrade(migrate_engine):
    """Add default quota class data into DB."""
    meta = MetaData()
    meta.bind = migrate_engine

    quota_classes = Table('quota_classes', meta, autoload=True)

    rows = quota_classes.count().\
        where(quota_classes.c.class_name == 'default').execute().scalar()

    # Do not add entries if there are already 'default' entries.  We don't
    # want to write over something the user added.
    if rows:
        return

    # Set default volumes
    qci = quota_classes.insert()
    qci.execute({'created_at': CREATED_AT,
                 'class_name': CLASS_NAME,
                 'resource': 'volumes',
                 'hard_limit': CONF.quota_volumes,
                 'deleted': False, })
    # Set default snapshots
    qci.execute({'created_at': CREATED_AT,
                 'class_name': CLASS_NAME,
                 'resource': 'snapshots',
                 'hard_limit': CONF.quota_snapshots,
                 'deleted': False, })
    # Set default gigabytes
    qci.execute({'created_at': CREATED_AT,
                 'class_name': CLASS_NAME,
                 'resource': 'gigabytes',
                 'hard_limit': CONF.quota_gigabytes,
                 'deleted': False, })



#    Copyright 2015 Intel Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from cinder import rpc
from cinder import test


class FakeAPI(rpc.RPCAPI):
    RPC_API_VERSION = '1.5'
    TOPIC = 'cinder-scheduler-topic'
    BINARY = 'cinder-scheduler'


class RPCAPITestCase(test.TestCase):
    """Tests RPCAPI mixin aggregating stuff related to RPC compatibility."""

    def setUp(self):
        super(RPCAPITestCase, self).setUp()
        # Reset cached version pins
        rpc.LAST_RPC_VERSIONS = {}
        rpc.LAST_OBJ_VERSIONS = {}

    @mock.patch('cinder.objects.Service.get_minimum_rpc_version',
                return_value='1.2')
    @mock.patch('cinder.objects.Service.get_minimum_obj_version',
                return_value='1.7')
    @mock.patch('cinder.rpc.get_client')
    def test_init(self, get_client, get_min_obj, get_min_rpc):
        def fake_get_client(target, version_cap, serializer):
            self.assertEqual(FakeAPI.TOPIC, target.topic)
            self.assertEqual(FakeAPI.RPC_API_VERSION, target.version)
            self.assertEqual('1.2', version_cap)
            self.assertEqual('1.7', serializer.version_cap)

        get_client.side_effect = fake_get_client
        FakeAPI()

    @mock.patch('cinder.objects.Service.get_minimum_rpc_version',
                return_value='liberty')
    @mock.patch('cinder.objects.Service.get_minimum_obj_version',
                return_value='liberty')
    @mock.patch('cinder.rpc.get_client')
    def test_init_liberty_caps(self, get_client, get_min_obj, get_min_rpc):
        def fake_get_client(target, version_cap, serializer):
            self.assertEqual(FakeAPI.TOPIC, target.topic)
            self.assertEqual(FakeAPI.RPC_API_VERSION, target.version)
            self.assertEqual(rpc.LIBERTY_RPC_VERSIONS[FakeAPI.BINARY],
                             version_cap)
            self.assertEqual('liberty', serializer.version_cap)

        get_client.side_effect = fake_get_client
        FakeAPI()

    @mock.patch('cinder.objects.Service.get_minimum_rpc_version')
    @mock.patch('cinder.objects.Service.get_minimum_obj_version')
    @mock.patch('cinder.rpc.get_client')
    @mock.patch('cinder.rpc.LAST_RPC_VERSIONS', {'cinder-scheduler': '1.4'})
    @mock.patch('cinder.rpc.LAST_OBJ_VERSIONS', {'cinder-scheduler': '1.3'})
    def test_init_cached_caps(self, get_client, get_min_obj, get_min_rpc):
        def fake_get_client(target, version_cap, serializer):
            self.assertEqual(FakeAPI.TOPIC, target.topic)
            self.assertEqual(FakeAPI.RPC_API_VERSION, target.version)
            self.assertEqual('1.4', version_cap)
            self.assertEqual('1.3', serializer.version_cap)

        get_client.side_effect = fake_get_client
        FakeAPI()

        self.assertFalse(get_min_obj.called)
        self.assertFalse(get_min_rpc.called)
